''' Loader functions for navigation visualisation data. '''
import pickle

from bluesky import settings
from bluesky.tools import cachefile
from bluesky.ui.loadvisuals_txt import load_coastline_txt, load_aptsurface_txt


# Cache versions: increment these to the current date if the source data is updated
# or other reasons why the cache needs to be updated
coast_version = 'v20170101'
navdb_version = 'v20170101'
aptsurf_version = 'v20171116'

## Default settings
settings.set_variable_defaults(navdata_path='navdata')


def load_coastlines():
    ''' Load coastline data for gui. '''
    with cachefile.openfile('coastlines.p', coast_version) as cache:
        try:
            coastvertices = cache.load()
            coastindices = cache.load()
        except (pickle.PickleError, cachefile.CacheError) as e:
            print(e.args[0])
            coastvertices, coastindices = load_coastline_txt()
            cache.dump(coastvertices)
            cache.dump(coastindices)

    return coastvertices, coastindices


def load_aptsurface():
    ''' Load airport surface polygons for gui. '''
    with cachefile.openfile('aptsurface.p', aptsurf_version) as cache:
        try:
            vbuf_asphalt = cache.load()
            vbuf_concrete = cache.load()
            vbuf_runways = cache.load()
            vbuf_rwythr = cache.load()
            apt_ctr_lat = cache.load()
            apt_ctr_lon = cache.load()
            apt_indices = cache.load()
        except (pickle.PickleError, cachefile.CacheError) as e:
            print(e.args[0])
            vbuf_asphalt, vbuf_concrete, vbuf_runways, vbuf_rwythr, apt_ctr_lat, \
                apt_ctr_lon, apt_indices = load_aptsurface_txt()
            cache.dump(vbuf_asphalt)
            cache.dump(vbuf_concrete)
            cache.dump(vbuf_runways)
            cache.dump(vbuf_rwythr)
            cache.dump(apt_ctr_lat)
            cache.dump(apt_ctr_lon)
            cache.dump(apt_indices)

    return vbuf_asphalt, vbuf_concrete, vbuf_runways, vbuf_rwythr, \
        apt_ctr_lat, apt_ctr_lon, apt_indices



""" Simple turbulence implementation."""
import numpy as np
import bluesky as bs
from bluesky.tools.aero import Rearth
from bluesky.core import Entity


class Turbulence(Entity, replaceable=True):
    """ Simple turbulence implementation."""
    def __init__(self):
        super().__init__()
        self.active = False
        self.sd = np.array([])

    def reset(self):
        self.active = False
        self.SetStandards([0, 0.1, 0.1])

    def setnoise(self, flag):
        self.active = flag

    def SetStandards(self, s):
        self.sd = np.array(s) # m/s standard turbulence  (nonnegative)
        # in (horizontal flight direction, horizontal wing direction, vertical)
        self.sd = np.where(self.sd > 1e-6, self.sd, 1e-6)

    def update(self):
        if not self.active:
            return

        timescale=np.sqrt(bs.sim.simdt)
        # Horizontal flight direction
        turbhf=np.random.normal(0,self.sd[0]*timescale,bs.traf.ntraf) #[m]

        # Horizontal wing direction
        turbhw=np.random.normal(0,self.sd[1]*timescale,bs.traf.ntraf) #[m]

        # Vertical direction
        turbalt=np.random.normal(0,self.sd[2]*timescale,bs.traf.ntraf) #[m]

        trkrad=np.radians(bs.traf.trk)
        # Lateral, longitudinal direction
        turblat=np.cos(trkrad)*turbhf-np.sin(trkrad)*turbhw #[m]
        turblon=np.sin(trkrad)*turbhf+np.cos(trkrad)*turbhw #[m]

        # Update the aircraft locations
        bs.traf.alt = bs.traf.alt + turbalt
        bs.traf.lat = bs.traf.lat + np.degrees(turblat/Rearth)
        bs.traf.lon = bs.traf.lon + np.degrees(turblon/Rearth/bs.traf.coslat)



from pathlib import Path
from hatchling.builders.hooks.plugin.interface import BuildHookInterface

from setuptools import Distribution, Extension
from setuptools.command import build_ext

def get_numpy_include():
    import numpy
    return [numpy.get_include()]


class CustomHook(BuildHookInterface):
    def initialize(self, version, build_data):
        print(version, build_data)
        extensions = [
            Extension('bluesky.tools.geo._cgeo', ['bluesky/tools/geo/src_cpp/_cgeo.cpp']),
            Extension('bluesky.traffic.asas.cstatebased', ['bluesky/traffic/asas/src_cpp/cstatebased.cpp'], include_dirs=['bluesky/tools/geo/src_cpp'])]

        dist = Distribution(dict(name='extended', include_dirs=get_numpy_include(),
                                 ext_modules=extensions))
        dist.package_dir = "extended"
        cmd = build_ext.build_ext(dist)
        cmd.verbose = True  # type: ignore
        cmd.ensure_finalized()  # type: ignore
        cmd.run()
        buildpath = Path(cmd.build_lib)

        # Provide locations of compiled modules
        force_include = {(buildpath / cmd.get_ext_filename('bluesky.tools.geo._cgeo')).as_posix(): cmd.get_ext_filename('bluesky.tools.geo._cgeo'),
                         (buildpath / cmd.get_ext_filename('bluesky.traffic.asas.cstatebased')).as_posix(): cmd.get_ext_filename('bluesky.traffic.asas.cstatebased')}

        build_data['pure_python'] = False
        build_data['infer_tag'] = True
        build_data['force_include'].update(force_include)


from os import path
from collections import defaultdict
from PyQt6.QtCore import Qt, pyqtSlot
from PyQt6.QtWidgets import QVBoxLayout, QScrollArea, QGroupBox, QWidget, \
    QFormLayout, QLabel, QSpinBox, QCheckBox, QLineEdit, QHBoxLayout, \
        QTreeWidget, QTreeWidgetItem, QFrame, QPushButton, QLayout, QComboBox, \
        QListWidget, QListWidgetItem

import bluesky as bs
from bluesky.network.common import seqidx2id
from bluesky.network.subscriber import subscribe
import bluesky.network.sharedstate as ss
from bluesky.core import Base


def sel_palette(value, changed_fun):
    wid = QComboBox()

    palfiles = [path.basename(f) for f in (bs.resource(bs.settings.gfx_path) / 'palettes').glob('*')]
    wid.addItems(palfiles)
    wid.setCurrentText(value)
    wid.currentTextChanged.connect(changed_fun)
    return wid

def sel_perf(value, changed_fun):
    wid = QComboBox()
    wid.addItems(['openap', 'bada', 'legacy'])
    wid.setCurrentText(value)
    wid.currentTextChanged.connect(changed_fun)
    return wid


class PluginCheckList(QListWidget):
    def __init__(self, value, changed_fun, avail_plugins):
        super().__init__()
        self.changed_fun = changed_fun
        self.curvalue = {v.upper() for v in value}
        if avail_plugins:
            for name in avail_plugins:
                row = QListWidgetItem(name)
                row.name = name
                row.setFlags(row.flags()|Qt.ItemIsUserCheckable)
                row.setCheckState(Qt.Checked if name in self.curvalue else Qt.Unchecked)
                self.addItem(row)
        self.itemChanged.connect(self.onitemchanged)

    def onitemchanged(self, item):
        if item.checkState() & Qt.Checked:
            self.curvalue.add(item.name)
        else:
            self.curvalue -= {item.name}
        self.changed_fun(list(self.curvalue))

class SettingsWindow(QWidget, Base):
    customwids = {
        'colour_palette': sel_palette,
        'performance_model': sel_perf,
        'enabled_plugins': PluginCheckList
    }
    def __init__(self):
        super().__init__()
        self.resize(600, 500)
        self.populated = False
        self.maxservnum = 0
        self.servers = dict()
        self.changed = defaultdict(dict)
        self.setLayout(QVBoxLayout())
        self.layout().setSpacing(0)
        self.layout().setContentsMargins(1, 1, 1, 1)
        self.scrollarea = QScrollArea()
        self.scrollarea.layout = QVBoxLayout()
        self.layout().addWidget(self.scrollarea)
        bottombar = QWidget()
        bottombar.setLayout(QHBoxLayout())
        bottombar.layout().setContentsMargins(11, 1, 1, 1)
        self.changedlabel = QLabel()
        self.resetbtn = QPushButton('Reset')
        self.resetbtn.setFixedWidth(100)
        self.resetbtn.setEnabled(False)
        self.resetbtn.clicked.connect(self.btnclicked)
        self.savebtn = QPushButton('Save')
        self.savebtn.setFixedWidth(100)
        self.savebtn.setEnabled(False)
        self.savebtn.clicked.connect(self.btnclicked)

        bottombar.layout().addWidget(self.changedlabel)
        bottombar.layout().addWidget(self.resetbtn)
        bottombar.layout().addWidget(self.savebtn)
        self.layout().addWidget(bottombar)

        self.nodetree = QTreeWidget()
        self.nodetree.setFixedWidth(200)
        self.nodetree.setIndentation(0)
        self.nodetree.setColumnCount(2)
        self.nodetree.setStyleSheet('padding:0px')
        self.nodetree.setAttribute(Qt.WidgetAttribute.WA_MacShowFocusRect, False)
        self.nodetree.header().resizeSection(0, 130)
        self.nodetree.setHeaderHidden(True)
        self.nodetree.itemClicked.connect(self.nodetreeClicked)
        self.nodesettings = QWidget()
        self.nodesettings.setLayout(QVBoxLayout())
        container = QWidget()
        container.setLayout(self.scrollarea.layout)
        self.scrollarea.layout.setAlignment(Qt.AlignmentFlag.AlignTop|Qt.AlignmentFlag.AlignLeft)
        self.scrollarea.setWidget(container)
        self.scrollarea.setWidgetResizable(True)

        bs.net.node_added.connect(self.nodesChanged)
        # Subscribe to simulation settings SharedState
        subscribe('SIMSETTINGS')

    def show(self):
        if not self.populated:
            self.populate()
        super().show()

    def populate(self):
        top = bs.settings._settings_hierarchy['bluesky']
        netset = top['network']
        netitems = {**netset['discovery'], **netset['server']}
        net = self.make_settings_box('Network settings', netitems)
        self.scrollarea.layout.addWidget(net)

        pathitems = {n:v for n, v in bs.settings.__dict__.items() if 'path' in n}
        paths = self.make_settings_box('Paths', pathitems)
        self.scrollarea.layout.addWidget(paths)

        guiitems = {n:v for n, v in top['ui']['qtgl']['radarwidget'].items() if 'path' not in n}
        guiitems['colour_palette'] = bs.settings.colour_palette
        gui = self.make_settings_box('Gui settings', guiitems, target='gui')
        self.scrollarea.layout.addWidget(gui)

        sim = QGroupBox('Simulation settings')
        sim.setLayout(QHBoxLayout())
        self.scrollarea.layout.addWidget(sim)
        sim.layout().addWidget(self.nodetree)
        sim.layout().addWidget(self.nodesettings)


    def nodesChanged(self, node_id):
        server_id = node_id[:-1] + seqidx2id(0)
        server = self.servers.get(server_id)
        if not server:
            server = QTreeWidgetItem(self.nodetree)
            self.maxservnum += 1
            server.serv_num = self.maxservnum
            server.serv_id = server_id
            server.nodes = [i for i in bs.net.nodes if i.startswith(server_id[:-1])]
            hostname = 'This computer'# if serv_id == bs.net.get_hostid() else str(serv_id)
            f = server.font(0)
            f.setBold(True)
            server.setExpanded(True)
            server.setText(0, hostname)
            server.setText(1, f'(nodes: {len(server.nodes)})')
            self.servers[server_id] = server

    @pyqtSlot(QTreeWidgetItem, int)
    def nodetreeClicked(self, item, column):
        if item in self.servers.values():
            simsettings = dict()
            plugins = list()

            for node in item.nodes:
                store = ss.get(node, 'simsettings')
                simsettings.update(store.settings)
                plugins = store.plugins

            # First clear any old items in the node settings layout
            clear_layout(self.nodesettings.layout())

            top = simsettings['bluesky']
            traf = self.make_settings_box('Traffic', top['traffic'], target=item.serv_id)
            stack = self.make_settings_box(
                'Stack', top['stack'], target=item.serv_id)
            sim = self.make_settings_box(
                'Simulation', top['simulation'], maxdepth=0, target=item.serv_id)

            misc = self.make_settings_box('Misc', top['tools'], maxdepth=0, target=item.serv_id, avail_plugins=plugins)
            self.nodesettings.layout().addWidget(traf)
            self.nodesettings.layout().addWidget(stack)
            self.nodesettings.layout().addWidget(sim)
            self.nodesettings.layout().addWidget(misc)
            for name, plugin in simsettings.items():
                if name == 'bluesky':
                    continue
                pbox = self.make_settings_box(
                    name.capitalize(), plugin, target=item.serv_id)
                self.nodesettings.layout().addWidget(pbox)

    def add_row(self, box, name, value, depth=0, maxdepth=1, target='common', **kwargs):
        if isinstance(value, dict):
            if depth < maxdepth:
                line = QFrame()
                
                line.setFrameShape(QFrame.Shape.HLine)
                box.layout().addRow(QLabel(f'<b>{name}</b>'), line)
            for cname, cvalue in value.items():
                self.add_row(box, cname, cvalue, depth=depth+1, target=target, **kwargs)
            return
        if name in SettingsWindow.customwids:
            wid = SettingsWindow.customwids[name](
                value, self.input_changed, **kwargs)
        elif isinstance(value, bool):
            wid = QCheckBox('')
            wid.setChecked(value)
            wid.clicked.connect(self.input_changed)
        elif isinstance(value, int):
            wid = QSpinBox()
            wid.setMaximum(100000)
            wid.setValue(value)
            wid.setFixedWidth(300)
            wid.valueChanged.connect(self.input_changed)
        else:
            wid = QLineEdit(str(value))
            wid.setFixedWidth(300)
            wid.textEdited.connect(self.input_changed)
        wid.origvalue = value
        wid.name = name
        wid.target = target
        box.layout().addRow(QLabel(name), wid)

    def input_changed(self, value):
        wid = self.sender()
        # Store the changed value in a dict of all changed values
        self.changed[wid.target][wid.name] = value
        self.changedlabel.setText('Save the changes and restart BlueSky for the changes to take effect.')
        self.resetbtn.setEnabled(True)
        self.savebtn.setEnabled(True)

    def make_settings_box(self, name, items, maxdepth=1, target='common', **kwargs):
        box = QGroupBox(name)
        box.setLayout(QFormLayout())
        for rowname, value in items.items():
            self.add_row(box, rowname, value, maxdepth=maxdepth,
                         target=target, **kwargs)

        return box

    def reset(self, branch=None):
        if hasattr(branch, 'layout'):
            layout = getattr(branch, 'layout')
            layout = layout() if callable(layout) else layout
            if layout is not branch and layout is not None:
                self.reset(layout)
                return
        if isinstance(branch, QLayout):
            for i in range(branch.count()):
                self.reset(branch.itemAt(i).widget())
        elif isinstance(branch, QCheckBox):
            branch.setChecked(branch.origvalue)
        elif isinstance(branch, QSpinBox):
            branch.setValue(branch.origvalue)
        elif isinstance(branch, QLineEdit):
            branch.setText(str(branch.origvalue))

    def save(self):
        common = self.changed.pop('common', {})
        localchanges = {**self.changed.pop('gui', {}), **self.changed.pop(bs.net.server_id, {}), **common}
        remotechanges = dict(self.changed)
        self.changed.clear()

        # TODO: send possible changes to other servers

        # Save local settings
        success, msg = bs.settings.save(changes=localchanges)
        # bs.scr.echo(msg)

    def btnclicked(self):
        self.resetbtn.setEnabled(False)
        self.savebtn.setEnabled(False)
        self.changedlabel.setText('')
        if self.sender() is self.savebtn:
            self.save()
        elif self.sender() is self.resetbtn:
            self.reset(self.scrollarea.layout)

    def hideEvent(self, event):
        pass

    def showEvent(self, event):
        pass

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Escape:
            self.close()

def clear_layout(layout):
    while layout.count():
        child = layout.takeAt(0)
        if child.widget():
            child.widget().deleteLater()



''' Conflict resolution based on the SSD algorithm. '''
from bluesky.traffic.asas import ConflictResolution
from bluesky.tools import geo
from bluesky.tools.aero import nm
import numpy as np
# Try to import pyclipper
try:
    import pyclipper
except ImportError:
    print("Could not import pyclipper, RESO SSD will not function")


# TODO: not completely migrated yet to class-based implementation


def init_plugin():

    # Addtional initilisation code

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'SSD',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim'
    }

    return config


class SSD(ConflictResolution):
    def setprio(self, flag=None, priocode=''):
        '''Set the prio switch and the type of prio '''
        if flag is None:
            return True, "PRIORULES [ON/OFF] [PRIOCODE]" + \
                            "\nAvailable priority codes: " + \
                            "\n     RS1:  Shortest way out" + \
                            "\n     RS2:  Clockwise turning" + \
                            "\n     RS3:  Heading first, RS1 second" + \
                            "\n     RS4:  Speed first, RS1 second" + \
                            "\n     RS5:  Shortest from target" + \
                            "\n     RS6:  Rules of the air" + \
                            "\n     RS7:  Sequential RS1" + \
                            "\n     RS8:  Sequential RS5" + \
                            "\n     RS9:  Counterclockwise turning" + \
                            "\nPriority is currently " + ("ON" if self.swprio else "OFF") + \
                            "\nPriority code is currently: " + \
                str(self.priocode)
        options = ["RS1", "RS2", "RS3", "RS4",
                   "RS5", "RS6", "RS7", "RS8", "RS9"]
        if priocode not in options:
            return False, "Priority code Not Understood. Available Options: " + str(options)
        return super().setprio(flag, priocode)

    def loaded_pyclipper():
        """ Return true if pyclipper is successfully loaded """
        import sys
        return "pyclipper" in sys.modules

    def detect(asas, traf):
        """ Detect all current conflicts """

        # Check if ASAS is ON first!
        if not asas.swasas:
            return

        # Construct the SSD
        constructSSD(asas, traf)


    def resolve(self, conf, ownship, intruder):
        # Initialize SSD variables with ntraf
        self.initializeSSD(conf, ownship.ntraf)

        # Construct the SSD (twice in sequential methods)
        if self.priocode == "RS7":
            self.constructSSD(conf, ownship, "RS1")
        if self.priocode == "RS8":
            self.constructSSD(conf, ownship, "RS5")
        self.constructSSD(conf, ownship, self.priocode)

        # Get resolved speed-vector
        self.calculate_resolution(conf, ownship)

        # Now assign resolutions to variables in the ASAS class
        # Start with current states, need a copy, otherwise it changes traf!
        newtrack = np.copy(ownship.hdg)
        newgs = np.copy(ownship.gs)
        # Calculate new track and speed
        # No need to cap the speeds, since SSD implicitly caps
        new_trk = np.arctan2(conf.asase, conf.asasn) * 180 / np.pi
        new_gs = np.sqrt(conf.asase ** 2 + conf.asasn ** 2)

        # Sometimes an aircraft is in conflict, but no solutions could be found
        # In that case it is assigned 0 by ASAS, but needs to handled
        asas_cmd = np.logical_and(conf.inconf, new_gs > 0)

        # Assign new track and speed for those that are in conflict
        newtrack[asas_cmd] = new_trk[asas_cmd]
        newgs[asas_cmd] = new_gs[asas_cmd]
        # Not needed as it is a 2D-implementation...
        newvs = ownship.vs

        # Cap the velocity
        newgscapped = np.maximum(ownship.perf.vmin, np.minimum(ownship.perf.vmax, newgs))

        alt = ownship.selalt

        return newtrack, newgscapped, newvs, alt


    def initializeSSD(self, conf, ntraf):
        """ Initialize variables for SSD """
        # Need to do it here, since ASAS.reset doesn't know ntraf
        conf.FRV = [None] * ntraf
        conf.ARV = [None] * ntraf
        # For calculation purposes
        conf.ARV_calc = [None] * ntraf
        conf.inrange = [None] * ntraf
        # asas.inconf       = np.zeros(ntraf, dtype=bool)
        # Index 2 for sequential solutions (RS7, RS8)
        conf.ARV_calc2 = [None] * ntraf
        conf.inrange2 = [None] * ntraf
        conf.inconf2 = np.zeros(ntraf, dtype=bool)
        # Stores resolution vector, also used in visualization
        conf.asasn = np.zeros(ntraf, dtype=np.float32)
        conf.asase = np.zeros(ntraf, dtype=np.float32)
        # Area calculation
        conf.FRV_area = np.zeros(ntraf, dtype=np.float32)
        conf.ARV_area = np.zeros(ntraf, dtype=np.float32)
        conf.ap_free = np.ones(ntraf, dtype=bool)

        # asas is an object of the ASAS class defined in asas.py


    def constructSSD(self, conf, ownship, priocode="RS1"):
        """ Calculates the FRV and ARV of the SSD """
        N = 0
        # Parameters
        N_angle = 180  # [-] Number of points on circle (discretization)
        hsep = conf.rpz  # [m] Horizontal separation (5 NM)
        margin = self.resofach  # [-] Safety margin for evasion
        hsepm = hsep * margin  # [m] Horizontal separation with safety margin
        alpham = 0.4999 * np.pi  # [rad] Maximum half-angle for VO
        betalos = np.pi / 4  # [rad] Minimum divertion angle for LOS (45 deg seems optimal)
        adsbmax = 65. * nm  # [m] Maximum ADS-B range
        beta = np.pi / 4 + betalos / 2
        if priocode == "RS7" or priocode == "RS8":
            adsbmax /= 2

        # Relevant info from traf
        gsnorth = ownship.gsnorth
        gseast = ownship.gseast
        lat = ownship.lat
        lon = ownship.lon
        ntraf = ownship.ntraf
        hdg = ownship.hdg
        gs_ap = ownship.ap.tas
        hdg_ap = ownship.ap.trk
        apnorth = np.cos(hdg_ap / 180 * np.pi) * gs_ap
        apeast = np.sin(hdg_ap / 180 * np.pi) * gs_ap

        # Local variables, will be put into asas later
        FRV_loc = [None] * ownship.ntraf
        ARV_loc = [None] * ownship.ntraf
        # For calculation purposes
        ARV_calc_loc = [None] * ownship.ntraf
        FRV_area_loc = np.zeros(ownship.ntraf, dtype=np.float32)
        ARV_area_loc = np.zeros(ownship.ntraf, dtype=np.float32)

        # # Use velocity limits for the ring-shaped part of the SSD
        # Discretize the circles using points on circle
        angles = np.arange(0, 2 * np.pi, 2 * np.pi / N_angle)
        # Put points of unit-circle in a (180x2)-array (CW)
        xyc = np.transpose(np.reshape(np.concatenate((np.sin(angles), np.cos(angles))), (2, N_angle)))

        # If no traffic
        if ntraf == 0:
            return

        # # If only one aircraft
        # elif ntraf == 1:
        #     # Map them into the format ARV wants. Outercircle CCW, innercircle CW
        #     ARV_loc[0] = circle_lst
        #     FRV_loc[0] = []
        #     ARV_calc_loc[0] = ARV_loc[0]
        #     # Calculate areas and store in asas
        #     FRV_area_loc[0] = 0
        #     ARV_area_loc[0] = np.pi * (vmax ** 2 - vmin ** 2)
        #     return

        # Function qdrdist_matrix needs 4 vectors as input (lat1,lon1,lat2,lon2)
        # To be efficient, calculate all qdr and dist in one function call
        # Example with ntraf = 5:   ind1 = [0,0,0,0,1,1,1,2,2,3]
        #                           ind2 = [1,2,3,4,2,3,4,3,4,4]
        # This way the qdrdist is only calculated once between every aircraft
        # To get all combinations, use this function to get the indices
        ind1, ind2 = self.qdrdist_matrix_indices(ntraf)
        # Get absolute bearing [deg] and distance [nm]
        # Not sure abs/rel, but qdr is defined from [-180,180] deg, w.r.t. North
        [qdr, dist] = geo.qdrdist_matrix(lat[ind1], lon[ind1], lat[ind2], lon[ind2])
        # Put result of function from matrix to ndarray
        qdr = np.reshape(np.array(qdr), np.shape(ind1))
        dist = np.reshape(np.array(dist), np.shape(ind1))
        # SI-units from [deg] to [rad]
        qdr = np.deg2rad(qdr)
        # Get distance from [nm] to [m]
        dist = dist * nm

        # In LoS the VO can't be defined, act as if dist is on edge
        dist[dist < hsepm] = hsepm

        # Calculate vertices of Velocity Obstacle (CCW)
        # These are still in relative velocity space, see derivation in appendix
        # Half-angle of the Velocity obstacle [rad]
        # Include safety margin
        alpha = np.arcsin(hsepm / dist)
        # Limit half-angle alpha to 89.982 deg. Ensures that VO can be constructed
        alpha[alpha > alpham] = alpham
        # Relevant sin/cos/tan
        sinqdr = np.sin(qdr)
        cosqdr = np.cos(qdr)
        tanalpha = np.tan(alpha)
        cosqdrtanalpha = cosqdr * tanalpha
        sinqdrtanalpha = sinqdr * tanalpha

        # Consider every aircraft
        for i in range(ntraf):
            # Calculate SSD only for aircraft in conflict (See formulas appendix)
            if conf.inconf[i]:

                vmin = ownship.perf.vmin[i]
                vmax = ownship.perf.vmax[i]

                # in the first time step, ASAS runs before perf, which means that his value will be zero
                # and the SSD cannot be constructed
                if vmin == vmax == 0:
                    continue

                # Map them into the format pyclipper wants. Outercircle CCW, innercircle CW
                circle_tup = (tuple(map(tuple, np.flipud(xyc * vmax))), tuple(map(tuple, xyc * vmin)))
                circle_lst = [list(map(list, np.flipud(xyc * vmax))), list(map(list, xyc * vmin))]

                # Relevant x1,y1,x2,y2 (x0 and y0 are zero in relative velocity space)
                x1 = (sinqdr + cosqdrtanalpha) * 2 * vmax
                x2 = (sinqdr - cosqdrtanalpha) * 2 * vmax
                y1 = (cosqdr - sinqdrtanalpha) * 2 * vmax
                y2 = (cosqdr + sinqdrtanalpha) * 2 * vmax

                # SSD for aircraft i
                # Get indices that belong to aircraft i
                ind = np.where(np.logical_or(ind1 == i, ind2 == i))[0]
                # Check whether there are any aircraft in the vicinity
                if len(ind) == 0:
                    # No aircraft in the vicinity
                    # Map them into the format ARV wants. Outercircle CCW, innercircle CW
                    ARV_loc[i] = circle_lst
                    FRV_loc[i] = []
                    ARV_calc_loc[i] = ARV_loc[i]
                    # Calculate areas and store in asas
                    FRV_area_loc[i] = 0
                    ARV_area_loc[i] = np.pi * (vmax ** 2 - vmin ** 2)
                else:
                    # The i's of the other aircraft
                    i_other = np.delete(np.arange(0, ntraf), i)
                    # Aircraft that are within ADS-B range
                    ac_adsb = np.where(dist[ind] < adsbmax)[0]
                    # Now account for ADS-B range in indices of other aircraft (i_other)
                    ind = ind[ac_adsb]
                    i_other = i_other[ac_adsb]
                    if not priocode == "RS7" and not priocode == "RS8":
                        # Put it in class-object (not for RS7 and RS8)
                        conf.inrange[i] = i_other
                    else:
                        conf.inrange2[i] = i_other
                    # VO from 2 to 1 is mirror of 1 to 2. Only 1 to 2 can be constructed in
                    # this manner, so need a correction vector that will mirror the VO
                    fix = np.ones(np.shape(i_other))
                    fix[i_other < i] = -1
                    # Relative bearing [deg] from [-180,180]
                    # (less required conversions than rad in RotA)
                    fix_ang = np.zeros(np.shape(i_other))
                    fix_ang[i_other < i] = 180.

                    # Get vertices in an x- and y-array of size (ntraf-1)*3x1
                    x = np.concatenate((gseast[i_other],
                                        x1[ind] * fix + gseast[i_other],
                                        x2[ind] * fix + gseast[i_other]))
                    y = np.concatenate((gsnorth[i_other],
                                        y1[ind] * fix + gsnorth[i_other],
                                        y2[ind] * fix + gsnorth[i_other]))
                    # Reshape [(ntraf-1)x3] and put arrays in one array [(ntraf-1)x3x2]
                    x = np.transpose(x.reshape(3, np.shape(i_other)[0]))
                    y = np.transpose(y.reshape(3, np.shape(i_other)[0]))
                    xy = np.dstack((x, y))

                    # Make a clipper object
                    pc = pyclipper.Pyclipper()
                    # Add circles (ring-shape) to clipper as subject
                    pc.AddPaths(pyclipper.scale_to_clipper(circle_tup), pyclipper.PT_SUBJECT, True)

                    # Extra stuff needed for RotA
                    if priocode == "RS6":
                        # Make another clipper object for RotA
                        pc_rota = pyclipper.Pyclipper()
                        pc_rota.AddPaths(pyclipper.scale_to_clipper(circle_tup), pyclipper.PT_SUBJECT, True)
                        # Bearing calculations from own view and other view
                        brg_own = np.mod((np.rad2deg(qdr[ind]) + fix_ang - hdg[i]) + 540., 360.) - 180.
                        brg_other = np.mod((np.rad2deg(qdr[ind]) + 180. - fix_ang - hdg[i_other]) + 540., 360.) - 180.

                    # Add each other other aircraft to clipper as clip
                    for j in range(np.shape(i_other)[0]):
                        ## Debug prints
                        ## print(traf.id[i] + " - " + traf.id[i_other[j]])
                        ## print(dist[ind[j]])
                        # Scale VO when not in LOS
                        if dist[ind[j]] > hsepm:
                            # Normally VO shall be added of this other a/c
                            VO = pyclipper.scale_to_clipper(tuple(map(tuple, xy[j, :, :])))
                        else:
                            # Pair is in LOS, instead of triangular VO, use darttip
                            # Check if bearing should be mirrored
                            if i_other[j] < i:
                                qdr_los = qdr[ind[j]] + np.pi
                            else:
                                qdr_los = qdr[ind[j]]
                            # Length of inner-leg of darttip
                            leg = 1.1 * vmax / np.cos(beta) * np.array([1, 1, 1, 0])
                            # Angles of darttip
                            angles_los = np.array([qdr_los + 2 * beta, qdr_los, qdr_los - 2 * beta, 0.])
                            # Calculate coordinates (CCW)
                            x_los = leg * np.sin(angles_los)
                            y_los = leg * np.cos(angles_los)
                            # Put in array of correct format
                            xy_los = np.vstack((x_los, y_los)).T
                            # Scale darttip
                            VO = pyclipper.scale_to_clipper(tuple(map(tuple, xy_los)))
                        # Add scaled VO to clipper
                        pc.AddPath(VO, pyclipper.PT_CLIP, True)
                        # For RotA it is possible to ignore
                        if priocode == "RS6":
                            if brg_own[j] >= -20. and brg_own[j] <= 110.:
                                # Head-on or converging from right
                                pc_rota.AddPath(VO, pyclipper.PT_CLIP, True)
                            elif brg_other[j] <= -110. or brg_other[j] >= 110.:
                                # In overtaking position
                                pc_rota.AddPath(VO, pyclipper.PT_CLIP, True)
                        # Detect conflicts for smaller layer in RS7 and RS8
                        if priocode == "RS7" or priocode == "RS8":
                            if pyclipper.PointInPolygon(pyclipper.scale_to_clipper((gseast[i], gsnorth[i])), VO):
                                conf.inconf2[i] = True
                        if priocode == "RS5":
                            if pyclipper.PointInPolygon(pyclipper.scale_to_clipper((apeast[i], apnorth[i])), VO):
                                conf.ap_free[i] = False

                    # Execute clipper command
                    FRV = pyclipper.scale_from_clipper(
                        pc.Execute(pyclipper.CT_INTERSECTION, pyclipper.PFT_NONZERO, pyclipper.PFT_NONZERO))

                    ARV = pc.Execute(pyclipper.CT_DIFFERENCE, pyclipper.PFT_NONZERO, pyclipper.PFT_NONZERO)

                    if not priocode == "RS1" and not priocode == "RS5" and not priocode == "RS7" and not priocode == "RS8":
                        # Make another clipper object for extra intersections
                        pc2 = pyclipper.Pyclipper()
                        # When using RotA clip with pc_rota
                        if priocode == "RS6":
                            # Calculate ARV for RotA
                            ARV_rota = pc_rota.Execute(pyclipper.CT_DIFFERENCE, pyclipper.PFT_NONZERO,
                                                       pyclipper.PFT_NONZERO)
                            if len(ARV_rota) > 0:
                                pc2.AddPaths(ARV_rota, pyclipper.PT_CLIP, True)
                        else:
                            # Put the ARV in there, make sure it's not empty
                            if len(ARV) > 0:
                                pc2.AddPaths(ARV, pyclipper.PT_CLIP, True)

                    # Scale back
                    ARV = pyclipper.scale_from_clipper(ARV)

                    # Check if ARV or FRV is empty
                    if len(ARV) == 0:
                        # No aircraft in the vicinity
                        # Map them into the format ARV wants. Outercircle CCW, innercircle CW
                        ARV_loc[i] = []
                        FRV_loc[i] = circle_lst
                        ARV_calc_loc[i] = []
                        # Calculate areas and store in asas
                        FRV_area_loc[i] = np.pi * (vmax ** 2 - vmin ** 2)
                        ARV_area_loc[i] = 0
                    elif len(FRV) == 0:
                        # Should not happen with one a/c or no other a/c in the vicinity.
                        # These are handled earlier. Happens when RotA has removed all
                        # Map them into the format ARV wants. Outercircle CCW, innercircle CW
                        ARV_loc[i] = circle_lst
                        FRV_loc[i] = []
                        ARV_calc_loc[i] = circle_lst
                        # Calculate areas and store in asas
                        FRV_area_loc[i] = 0
                        ARV_area_loc[i] = np.pi * (vmax ** 2 - vmin ** 2)
                    else:
                        # Check multi exteriors, if this layer is not a list, it means it has no exteriors
                        # In that case, make it a list, such that its format is consistent with further code
                        if not type(FRV[0][0]) == list:
                            FRV = [FRV]
                        if not type(ARV[0][0]) == list:
                            ARV = [ARV]
                        # Store in asas
                        FRV_loc[i] = FRV
                        ARV_loc[i] = ARV
                        # Calculate areas and store in asas
                        FRV_area_loc[i] = self.area(FRV)
                        ARV_area_loc[i] = self.area(ARV)

                        # For resolution purposes sometimes extra intersections are wanted
                        if priocode == "RS2" or priocode == "RS9" or priocode == "RS6" or priocode == "RS3" or priocode == "RS4":
                            # Make a box that covers right or left of SSD
                            own_hdg = hdg[i] * np.pi / 180
                            # Efficient calculation of box, see notes
                            if priocode == "RS2" or priocode == "RS6":
                                # CW or right-turning
                                sin_table = np.array([[1, 0], [-1, 0], [-1, -1], [1, -1]], dtype=np.float64)
                                cos_table = np.array([[0, 1], [0, -1], [1, -1], [1, 1]], dtype=np.float64)
                            elif priocode == "RS9":
                                # CCW or left-turning
                                sin_table = np.array([[1, 0], [1, 1], [-1, 1], [-1, 0]], dtype=np.float64)
                                cos_table = np.array([[0, 1], [-1, 1], [-1, -1], [0, -1]], dtype=np.float64)
                            # Overlay a part of the full SSD
                            if priocode == "RS2" or priocode == "RS9" or priocode == "RS6":
                                # Normalized coordinates of box
                                xyp = np.sin(own_hdg) * sin_table + np.cos(own_hdg) * cos_table
                                # Scale with vmax (and some factor) and put in tuple
                                part = pyclipper.scale_to_clipper(tuple(map(tuple, 1.1 * vmax * xyp)))
                                pc2.AddPath(part, pyclipper.PT_SUBJECT, True)
                            elif priocode == "RS3":
                                # Small ring
                                xyp = (tuple(map(tuple, np.flipud(xyc * min(vmax, gs_ap[i] + 0.1)))),
                                       tuple(map(tuple, xyc * max(vmin, gs_ap[i] - 0.1))))
                                part = pyclipper.scale_to_clipper(xyp)
                                pc2.AddPaths(part, pyclipper.PT_SUBJECT, True)
                            elif priocode == "RS4":
                                hdg_sel = hdg[i] * np.pi / 180
                                xyp = np.array([[np.sin(hdg_sel - 0.0087), np.cos(hdg_sel - 0.0087)],
                                                [0, 0],
                                                [np.sin(hdg_sel + 0.0087), np.cos(hdg_sel + 0.0087)]],
                                               dtype=np.float64)
                                part = pyclipper.scale_to_clipper(tuple(map(tuple, 1.1 * vmax * xyp)))
                                pc2.AddPath(part, pyclipper.PT_SUBJECT, True)
                            # Execute clipper command
                            ARV_calc = pyclipper.scale_from_clipper(
                                pc2.Execute(pyclipper.CT_INTERSECTION, pyclipper.PFT_NONZERO, pyclipper.PFT_NONZERO))
                            N += 1
                            # If no smaller ARV is found, take the full ARV
                            if len(ARV_calc) == 0:
                                ARV_calc = ARV
                            # Check multi exteriors, if this layer is not a list, it means it has no exteriors
                            # In that case, make it a list, such that its format is consistent with further code
                            if not type(ARV_calc[0][0]) == list:
                                ARV_calc = [ARV_calc]
                        # Shortest way out prio, so use full SSD (ARV_calc = ARV)
                        else:
                            ARV_calc = ARV
                        # Update calculatable ARV for resolutions
                        ARV_calc_loc[i] = ARV_calc

        # If sequential approach, the local should go elsewhere
        if not priocode == "RS7" and not priocode == "RS8":
            conf.FRV = FRV_loc
            conf.ARV = ARV_loc
            conf.ARV_calc = ARV_calc_loc
            conf.FRV_area = FRV_area_loc
            conf.ARV_area = ARV_area_loc
        else:
            conf.ARV_calc2 = ARV_calc_loc
        return


    def calculate_resolution(self, conf, ownship):
        """ Calculates closest conflict-free point according to ruleset """
        # It's just linalg, however credits to: http://stackoverflow.com/a/1501725
        # Variables
        ARV = conf.ARV_calc
        if self.priocode == "RS7" or self.priocode == "RS8":
            ARV2 = conf.ARV_calc2
        # Select AP-setting as reference point for closest to target rulesets
        if self.priocode == "RS5" or self.priocode == "RS8":
            gsnorth = np.cos(ownship.ap.trk / 180 * np.pi) * ownship.ap.tas
            gseast = np.sin(ownship.ap.trk / 180 * np.pi) * ownship.ap.tas
        else:
            gsnorth = ownship.gsnorth
            gseast = ownship.gseast
        ntraf = ownship.ntraf

        # Loop through SSDs of all aircraft
        for i in range(ntraf):
            # Only those that are in conflict need to resolve
            if conf.inconf[i] and ARV[i] is not None and len(ARV[i]) > 0:
                # First check if AP-setting is free
                if conf.ap_free[i] and self.priocode == "RS5":
                    conf.asase[i] = gseast[i]
                    conf.asasn[i] = gsnorth[i]
                else:
                    # Loop through all exteriors and append. Afterwards concatenate
                    p = []
                    q = []
                    for j in range(len(ARV[i])):
                        p.append(np.array(ARV[i][j]))
                        q.append(np.diff(np.row_stack((p[j], p[j][0])), axis=0))
                    p = np.concatenate(p)
                    q = np.concatenate(q)
                    # Calculate squared distance between edges
                    l2 = np.sum(q ** 2, axis=1)
                    # Catch l2 == 0 (exception)
                    same = l2 < 1e-8
                    l2[same] = 1.
                    # Calc t
                    t = np.sum((np.array([gseast[i], gsnorth[i]]) - p) * q, axis=1) / l2
                    # Speed of boolean indices only slightly faster (negligible)
                    # t must be limited between 0 and 1
                    t = np.clip(t, 0., 1.)
                    t[same] = 0.
                    # Calculate closest point to each edge
                    x1 = p[:, 0] + t * q[:, 0]
                    y1 = p[:, 1] + t * q[:, 1]
                    # Get distance squared
                    d2 = (x1 - gseast[i]) ** 2 + (y1 - gsnorth[i]) ** 2
                    # Sort distance
                    ind = np.argsort(d2)
                    x1 = x1[ind]
                    y1 = y1[ind]

                    if self.priocode == "RS7" or self.priocode == "RS8" and conf.inconf2[i]:
                        # Loop through all exteriors and append. Afterwards concatenate
                        p = []
                        q = []
                        for j in range(len(ARV2[i])):
                            p.append(np.array(ARV2[i][j]))
                            q.append(np.diff(np.row_stack((p[j], p[j][0])), axis=0))
                        p = np.concatenate(p)
                        q = np.concatenate(q)
                        # Calculate squared distance between edges
                        l2 = np.sum(q ** 2, axis=1)
                        # Catch l2 == 0 (exception)
                        same = l2 < 1e-8
                        l2[same] = 1.
                        # Calc t
                        t = np.sum((np.array([gseast[i], gsnorth[i]]) - p) * q, axis=1) / l2
                        # Speed of boolean indices only slightly faster (negligible)
                        # t must be limited between 0 and 1
                        t = np.clip(t, 0., 1.)
                        t[same] = 0.
                        # Calculate closest point to each edge
                        x2 = p[:, 0] + t * q[:, 0]
                        y2 = p[:, 1] + t * q[:, 1]
                        # Get distance squared
                        d2 = (x2 - gseast[i]) ** 2 + (y2 - gsnorth[i]) ** 2
                        # Sort distance
                        ind = np.argsort(d2)
                        x2 = x2[ind]
                        y2 = y2[ind]
                        d2 = d2[ind]

                    # Store result in conf
                    if not conf.inconf2[i] or not self.priocode == "RS7" and not self.priocode == "RS8":
                        conf.asase[i] = x1[0]
                        conf.asasn[i] = y1[0]
                    else:
                        # Sequential method, check if both result in very similar resolutions
                        if (x1[0] - x2[0]) * (x1[0] - x2[0]) + (x1[0] - x2[0]) * (x1[0] - x2[0]) < 1:
                            # In that case take the full layer
                            conf.asase[i] = x1[0]
                            conf.asasn[i] = y1[0]
                        else:
                            # In that case take the partial layer solution and see which
                            # results in lower TLOS
                            conf.asase[i] = x1[0]
                            conf.asasn[i] = y1[0]
                            # dv2 for the RS1-solution
                            dist12 = (x1[0] - gseast[i]) ** 2 + (y1[0] - gsnorth[i]) ** 2
                            # distances for the partial layer solution stored in d2
                            ind = d2 < dist12
                            if sum(ind) == 1:
                                conf.asase[i] = x2[0]
                                conf.asasn[i] = y2[0]
                            elif sum(ind) > 1:
                                x2 = x2[ind]
                                y2 = y2[ind]
                                # Get solution with minimum TLOS
                                i_other = conf.inrange[i]
                                idx = self.minTLOS(conf, ownship, i, i_other, x1, y1, x2, y2)
                                # Get solution with maximum TLOS
                                conf.asase[i] = x2[idx]
                                conf.asasn[i] = y2[idx]
                            else:
                                # This case should not happen...
                                conf.asase[i] = x1[0]
                                conf.asasn[i] = y1[0]

            # Those that are not in conflict will be assigned zeros
            # Or those that have no solutions (full ARV)
            else:
                conf.asase[i] = 0.
                conf.asasn[i] = 0.


    def area(self, vset):
        """ This function calculates the area of the set of FRV or ARV """
        # Initialize A as it could be calculated iteratively
        A = 0
        # Check multiple exteriors
        if type(vset[0][0]) == list:
            # Calc every exterior separately
            for i in range(len(vset)):
                A += pyclipper.scale_from_clipper(
                    pyclipper.scale_from_clipper(pyclipper.Area(pyclipper.scale_to_clipper(vset[i]))))
        else:
            # Single exterior
            A = pyclipper.scale_from_clipper(
                pyclipper.scale_from_clipper(pyclipper.Area(pyclipper.scale_to_clipper(vset))))
        return A


    def qdrdist_matrix_indices(self, ntraf):
        """ This function gives the indices that can be used in the lon/lat-vectors """
        # The indices will be n*(n-1)/2 long
        # Only works for n >= 2, which is logical...
        # This is faster than np.triu_indices :)
        tmp_range = np.arange(ntraf - 1, dtype=np.int32)
        ind1 = np.repeat(tmp_range, (tmp_range + 1)[::-1])
        ind2 = np.ones(ind1.shape[0], dtype=np.int32)
        inds = np.cumsum(tmp_range[1:][::-1] + 1)
        np.put(ind2, inds, np.arange(ntraf * -1 + 3, 1))
        ind2 = np.cumsum(ind2, out=ind2)
        return ind1, ind2


    def minTLOS(self, conf, ownship, i, i_other, x1, y1, x, y):
        """ This function calculates the aggregated TLOS for all resolution points """
        # Get speeds of other AC in range
        x_other = ownship.gseast[i_other]
        y_other = ownship.gsnorth[i_other]
        # Get relative bearing [deg] and distance [nm]
        qdr, dist = geo.qdrdist(ownship.lat[i], ownship.lon[i], ownship.lat[i_other], ownship.lon[i_other])
        # Convert to SI
        qdr = np.deg2rad(qdr)
        dist *= nm
        # For vectorization, store lengths as W and L
        W = np.shape(x)[0]
        L = np.shape(x_other)[0]
        # Relative speed-components
        du = np.dot(x_other.reshape((L, 1)), np.ones((1, W))) - np.dot(np.ones((L, 1)), x.reshape((1, W)))
        dv = np.dot(y_other.reshape((L, 1)), np.ones((1, W))) - np.dot(np.ones((L, 1)), y.reshape((1, W)))
        # Relative speed + zero check
        vrel2 = du * du + dv * dv
        vrel2 = np.where(np.abs(vrel2) < 1e-6, 1e-6, vrel2)  # limit lower absolute value
        # X and Y distance
        dx = np.dot(np.reshape(dist * np.sin(qdr), (L, 1)), np.ones((1, W)))
        dy = np.dot(np.reshape(dist * np.cos(qdr), (L, 1)), np.ones((1, W)))
        # Time to CPA
        tcpa = -(du * dx + dv * dy) / vrel2
        # CPA distance
        dcpa2 = np.square(np.dot(dist.reshape((L, 1)), np.ones((1, W)))) - np.square(tcpa) * vrel2
        # Calculate time to LOS
        R2 = conf.rpz * conf.rpz
        swhorconf = dcpa2 < R2
        dxinhor = np.sqrt(np.maximum(0, R2 - dcpa2))
        dtinhor = dxinhor / np.sqrt(vrel2)
        tinhor = np.where(swhorconf, tcpa - dtinhor, 0.)
        tinhor = np.where(tinhor > 0, tinhor, 1e6)
        # Get index of best solution
        idx = np.argmax(np.sum(tinhor, 0))

        return idx


""" Network functionality for BlueSky."""
import time
import socket
import threading
import bluesky as bs
import sys


def as_bytes(msg):
    """
    Encodes strings to bytes.
    """
    if sys.version_info.major == 3:
        return msg.encode('utf-8')
    else:
        return msg


class TcpSocket:
    """A TCP Client receving message from server, analysing the data, and """
    def __init__(self):
        self.buffer_size = 1024
        self.is_connected = False
        self.receiver_thread = threading.Thread(target=self.receiver)
        self.receiver_thread.daemon = True
        self.receiver_thread.start()

    def connectToHost(self, ip, port):
        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.setblocking(0)
            self.sock.settimeout(10)    # 10 second timeout
            self.sock.connect((ip, port))       # connecting
            self.is_connected = True
            print("Server connected. HOST: %s, PORT: %s" % (ip, port))
        except Exception as err:
            self.is_connected = False
            print("Connection Error: %s" % err)
            pass

    def disconnectFromHost(self):
        try:
            self.sock.close()
            self.is_connected = False
            print("Server disconnected.")
        except Exception as err:
            print("Disconnection Error: %s" % err)
            pass

    def isConnected(self):
        return self.is_connected

    def receiver(self):
        while True:
            if not self.is_connected:
                time.sleep(1.0)
                continue

            try:
                data = self.sock.recv(self.buffer_size)
                self.processData(data)
                time.sleep(0.1)
            except Exception as err:
                print("Receiver Error: %s" % err)
                time.sleep(1.0)

    def processData(self, data):
        # rewrite this function
        print("parsing data...")

    def numConnections(self):
        return None


class TcpServer:
    def __init__(self):
        pass

    def sendReply(self, event):
        pass

    def start(self):
        pass

    def processData(self, data, sender_id):
        pass



import numpy as np
from bluesky.tools.aero import nm, ft
from bluesky.traffic.asas import MVP
from bluesky import traf
# TODO: not completely migrated yet to class-based implementation


def init_plugin():

    # Addtional initilisation code

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'SWARM',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim'
    }

    return config


class Swarm(MVP):
    def __init__(self):
        super().__init__()
        self.rpzswarm = 7.5 * nm  # [m]
        self.dhswarm = 1500 * ft  # [m]
        self.Swarmweights = np.array([10, 3, 1])

    def resolve(self, conf, ownship, intruder):
        # Find matrix of neighbouring aircraft withing swarm distance
        qdrrad = np.radians(conf.qdr)
        dx = conf.dist * np.sin(qdrrad)
        dy = conf.dist * np.cos(qdrrad)

        dy = dy - np.eye(ownship.ntraf) * 1e9  # setting distance of A/C to itself to 0,\
        # correcting the distance from CASAS line 109

        dalt = ownship.alt.reshape((1, ownship.ntraf)) - ownship.alt.reshape((1, ownship.ntraf)).T
        close = np.logical_and(dx**2 + dy**2 < conf.rpzswarm**2,
                            np.abs(dalt) < self.dhswarm)

        trkdif = ownship.trk.reshape(1, ownship.ntraf) - ownship.trk.reshape(ownship.ntraf, 1)
        dtrk = (trkdif + 180) % 360 - 180
        samedirection = np.abs(dtrk) < 90

        selected = np.logical_and(close, samedirection)
        own = np.eye(ownship.ntraf, dtype='bool')

        Swarming = np.logical_or(selected, own)

        # First do conflict resolution following MVP
        newtrk, newgs, newvs, newalt = super().resolve(conf, ownship, intruder)

        # Find desired speed vector after Collision Avoidance or Autopilot
        ca_trk = np.where(conf.inconf, newtrk, ownship.ap.trk)
        ca_cas = np.where(conf.inconf, newgs, ownship.selspd)
        ca_vs = np.where(conf.inconf, newvs, ownship.selvs)

        # Add factor of Velocity Alignment to speed vector
        hspeed = np.ones((ownship.ntraf, ownship.ntraf)) * ownship.cas
        va_cas = np.average(hspeed, axis=1, weights=Swarming)

        vspeed = np.ones((ownship.ntraf, ownship.ntraf)) * ownship.vs
        va_vs = np.average(vspeed, axis=1, weights=Swarming)

        avgdtrk = np.average(dtrk, axis=1, weights=Swarming)
        va_trk = ownship.trk + avgdtrk

        # Add factor of Flock Centering to speed vector
        dxflock = dx + np.eye(ownship.ntraf) * asas.u / 100.
        dyflock = dy + np.eye(ownship.ntraf) * asas.v / 100.

        fc_dx = np.average(dxflock, axis=1, weights=Swarming)
        fc_dy = np.average(dyflock, axis=1, weights=Swarming)

        z = np.ones((traf.ntraf, traf.ntraf)) * traf.alt
        fc_dz = np.average(z, axis=1, weights=Swarming) - traf.alt

        fc_trk = np.degrees(np.arctan2(fc_dx, fc_dy))
        fc_cas = traf.cas
        ttoreach = np.sqrt(fc_dx**2 + fc_dy**2) / fc_cas
        fc_vs = np.where(ttoreach == 0, 0, fc_dz / ttoreach)

        # Find final Swarming directions
        trks = np.array([ca_trk, va_trk, fc_trk])
        cass = np.array([ca_cas, va_cas, fc_cas])
        vss = np.array([ca_vs, va_vs, fc_vs])

        trksrad = np.radians(trks)
        vxs = cass * np.sin(trksrad)
        vys = cass * np.cos(trksrad)

        Swarmvx = np.average(vxs, axis=0, weights=asas.Swarmweights)
        Swarmvy = np.average(vys, axis=0, weights=asas.Swarmweights)
        Swarmhdg = np.degrees(np.arctan2(Swarmvx, Swarmvy))
        Swarmcas = np.average(cass, axis=0, weights=asas.Swarmweights)
        Swarmvs = np.average(vss, axis=0, weights=asas.Swarmweights)

        # Cap the velocity
        Swarmcascapped = np.maximum(traf.perf.vmin, np.minimum(traf.perf.vmax, Swarmcas))
        # Assign Final Swarming directions to traffic
        asas.hdg = Swarmhdg
        asas.tas = Swarmcascapped
        asas.vs = Swarmvs
        asas.alt = np.sign(Swarmvs) * 1e5

        # Make sure that all aircraft follow these directions
        asas.active.fill(True)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from distutils.core import setup, Extension
import numpy as np

ext_modules = [Extension('cstatebased', sources=['cstatebased.cpp'])]

setup(name='cstatebased', version='1.0', include_dirs=[np.get_include(), '../../../tools/src_cpp'],
      ext_modules=ext_modules)



''' State-based conflict detection. '''
import numpy as np
from bluesky import stack
from bluesky.tools import geo
from bluesky.tools.aero import nm
from bluesky.traffic.asas import ConflictDetection


class StateBased(ConflictDetection):
    def detect(self, ownship, intruder, rpz, hpz, dtlookahead):
        ''' Conflict detection between ownship (traf) and intruder (traf/adsb).'''
        # Identity matrix of order ntraf: avoid ownship-ownship detected conflicts
        I = np.eye(ownship.ntraf)

        # Horizontal conflict ------------------------------------------------------

        # qdrlst is for [i,j] qdr from i to j, from perception of ADSB and own coordinates
        qdr, dist = geo.kwikqdrdist_matrix(np.asmatrix(ownship.lat), np.asmatrix(ownship.lon),
                                    np.asmatrix(intruder.lat), np.asmatrix(intruder.lon))

        # Convert back to array to allow element-wise array multiplications later on
        # Convert to meters and add large value to own/own pairs
        qdr = np.asarray(qdr)
        dist = np.asarray(dist) * nm + 1e9 * I

        # Calculate horizontal closest point of approach (CPA)
        qdrrad = np.radians(qdr)
        dx = dist * np.sin(qdrrad)  # is pos j rel to i
        dy = dist * np.cos(qdrrad)  # is pos j rel to i

        # Ownship track angle and speed
        owntrkrad = np.radians(ownship.trk)
        ownu = ownship.gs * np.sin(owntrkrad).reshape((1, ownship.ntraf))  # m/s
        ownv = ownship.gs * np.cos(owntrkrad).reshape((1, ownship.ntraf))  # m/s

        # Intruder track angle and speed
        inttrkrad = np.radians(intruder.trk)
        intu = intruder.gs * np.sin(inttrkrad).reshape((1, ownship.ntraf))  # m/s
        intv = intruder.gs * np.cos(inttrkrad).reshape((1, ownship.ntraf))  # m/s

        du = ownu - intu.T  # Speed du[i,j] is perceived eastern speed of i to j
        dv = ownv - intv.T  # Speed dv[i,j] is perceived northern speed of i to j

        dv2 = du * du + dv * dv
        dv2 = np.where(np.abs(dv2) < 1e-6, 1e-6, dv2)  # limit lower absolute value
        vrel = np.sqrt(dv2)

        tcpa = -(du * dx + dv * dy) / dv2 + 1e9 * I

        # Calculate distance^2 at CPA (minimum distance^2)
        dcpa2 = np.abs(dist * dist - tcpa * tcpa * dv2)

        # Check for horizontal conflict
        # RPZ can differ per aircraft, get the largest value per aircraft pair
        rpz = np.asarray(np.maximum(np.asmatrix(rpz), np.asmatrix(rpz).transpose()))
        R2 = rpz * rpz
        swhorconf = dcpa2 < R2  # conflict or not

        # Calculate times of entering and leaving horizontal conflict
        dxinhor = np.sqrt(np.maximum(0., R2 - dcpa2))  # half the distance travelled inzide zone
        dtinhor = dxinhor / vrel

        tinhor = np.where(swhorconf, tcpa - dtinhor, 1e8)  # Set very large if no conf
        touthor = np.where(swhorconf, tcpa + dtinhor, -1e8)  # set very large if no conf

        # Vertical conflict --------------------------------------------------------

        # Vertical crossing of disk (-dh,+dh)
        dalt = ownship.alt.reshape((1, ownship.ntraf)) - \
            intruder.alt.reshape((1, ownship.ntraf)).T  + 1e9 * I

        dvs = ownship.vs.reshape(1, ownship.ntraf) - \
            intruder.vs.reshape(1, ownship.ntraf).T
        dvs = np.where(np.abs(dvs) < 1e-6, 1e-6, dvs)  # prevent division by zero

        # Check for passing through each others zone
        # hPZ can differ per aircraft, get the largest value per aircraft pair
        hpz = np.asarray(np.maximum(np.asmatrix(hpz), np.asmatrix(hpz).transpose()))
        tcrosshi = (dalt + hpz) / -dvs
        tcrosslo = (dalt - hpz) / -dvs
        tinver = np.minimum(tcrosshi, tcrosslo)
        toutver = np.maximum(tcrosshi, tcrosslo)

        # Combine vertical and horizontal conflict----------------------------------
        tinconf = np.maximum(tinver, tinhor)
        toutconf = np.minimum(toutver, touthor)

        swconfl = np.array(swhorconf * (tinconf <= toutconf) * (toutconf > 0.0) *
                           np.asarray(tinconf < np.asmatrix(dtlookahead).T) * (1.0 - I), dtype=bool)

        # --------------------------------------------------------------------------
        # Update conflict lists
        # --------------------------------------------------------------------------
        # Ownship conflict flag and max tCPA
        inconf = np.any(swconfl, 1)
        tcpamax = np.max(tcpa * swconfl, 1)

        # Select conflicting pairs: each a/c gets their own record
        confpairs = [(ownship.id[i], ownship.id[j]) for i, j in zip(*np.where(swconfl))]
        swlos = (dist < rpz) * (np.abs(dalt) < hpz)
        lospairs = [(ownship.id[i], ownship.id[j]) for i, j in zip(*np.where(swlos))]

        return confpairs, lospairs, inconf, tcpamax, \
            qdr[swconfl], dist[swconfl], np.sqrt(dcpa2[swconfl]), \
                tcpa[swconfl], tinconf[swconfl]


try:
    from bluesky.traffic.asas import cstatebased


    class CStateBased(StateBased):
        def __init__(self):
            super().__init__()
            self.detect = cstatebased.detect

except ImportError:
    pass



""" ScreenIO is a screen proxy on the simulation side for the QTGL implementation of BlueSky."""
import time
import numpy as np

# Local imports
import bluesky as bs
from bluesky import stack
from bluesky.core import Entity
from bluesky.tools import aero
from bluesky.core.walltime import Timer
from bluesky.network.publisher import state_publisher, StatePublisher


# =========================================================================
# Settings
# =========================================================================
# Update rate of simulation info messages [Hz]
SIMINFO_RATE = 1

# Update rate of aircraft update messages [Hz]
ACUPDATE_RATE = 5


class ScreenIO(Entity):
    """Class within sim task which sends/receives data to/from GUI task"""

    pub_defwpt = StatePublisher('DEFWPT', collect=True)
    pub_route = StatePublisher('ROUTEDATA')

    # =========================================================================
    # Functions
    # =========================================================================
    def __init__(self):
        super().__init__()
        # Screen state defaults
        self.def_pan     = (0.0, 0.0)
        self.def_zoom    = 1.0
        self.route_all   = ""

        # Screen state overrides per client
        self.client_route = dict()

        # Dicts of custom aircraft and group colors
        self.custacclr = dict()
        self.custgrclr = dict()

        # Timing bookkeeping counters
        self.prevtime    = 0.0
        self.samplecount = 0
        self.prevcount   = 0

        # Output event timers
        self.slow_timer = Timer(1000 // SIMINFO_RATE)
        self.slow_timer.timeout.connect(self.send_siminfo)
        self.slow_timer.timeout.connect(self.send_route_data)

    def update(self):
        if bs.sim.state == bs.OP:
            self.samplecount += 1

    def reset(self):
        self.client_route = dict()
        self.route_all = ''
        self.custacclr = dict()
        self.custgrclr = dict()
        self.samplecount = 0
        self.prevcount   = 0
        self.prevtime    = 0.0

        self.def_pan = (0.0, 0.0)
        self.def_zoom = 1.0

    def echo(self, text='', flags=0):
        bs.net.send(b'ECHO', dict(text=text, flags=flags))

    def cmdline(self, text):
        bs.net.send(b'CMDLINE', text)

    def showroute(self, acid):
        ''' Toggle show route for this aircraft '''
        if not stack.sender():
            self.route_all = acid
            self.client_route.clear()
        else:
            self.client_route[stack.sender()] = acid
        return True

    def addnavwpt(self, name, lat, lon):
        ''' Add custom waypoint to visualization '''
        self.pub_defwpt.send_append(custwplbl=name, custwplat=lat, custwplon=lon)
        return True

    def removenavwpt(self, name):
        ''' Remove custom waypoint to visualization '''
        self.pub_defwpt.send_delete(custwplbl=name)
        return True

    def show_file_dialog(self):
        bs.net.send(b'SHOWDIALOG', dict(dialog='OPENFILE'))
        return ''

    def show_cmd_doc(self, cmd=''):
        bs.net.send(b'SHOWDIALOG', dict(dialog='DOC', args=cmd))

    @stack.commandgroup(annotations='txt,color', aliases=('COLOR', 'COL'))
    def colour(self, name, r, g, b):
        ''' Set custom color for aircraft or shape. '''
        if name in bs.traf.groups:
            groupmask = bs.traf.groups.groups[name]
            self.custgrclr[groupmask] = (r, g, b)
        elif name in bs.traf.id:
            self.custacclr[name] = (r, g, b)

    # =========================================================================
    # Slots
    # =========================================================================
    def send_siminfo(self):
        t  = time.time()
        dt = np.maximum(t - self.prevtime, 0.00001)  # avoid divide by 0
        speed = (self.samplecount - self.prevcount) / dt * bs.sim.simdt
        self.prevtime  = t
        self.prevcount = self.samplecount
        bs.net.send(b'SIMINFO', (speed, bs.sim.simdt, bs.sim.simt,
            str(bs.sim.utc.replace(microsecond=0)), bs.traf.ntraf, bs.sim.state, stack.get_scenname()))
        

    @state_publisher(topic='TRAILS', dt=1000 // SIMINFO_RATE, send_type='extend')
    def send_trails(self):
        # Trails, send only new line segments to be added
        if bs.traf.trails.active and len(bs.traf.trails.newlat0) > 0:
            data = dict(traillat0=bs.traf.trails.newlat0,
                        traillon0=bs.traf.trails.newlon0,
                        traillat1=bs.traf.trails.newlat1,
                        traillon1=bs.traf.trails.newlon1)
            bs.traf.trails.clearnew()
            return data

    @state_publisher(topic='ACDATA', dt=1000 // ACUPDATE_RATE)
    def send_aircraft_data(self):
        data = dict()
        data['simt']       = bs.sim.simt
        data['id']         = bs.traf.id
        data['lat']        = bs.traf.lat
        data['lon']        = bs.traf.lon
        data['alt']        = bs.traf.alt
        data['tas']        = bs.traf.tas
        data['cas']        = bs.traf.cas
        data['gs']         = bs.traf.gs
        data['ingroup']    = bs.traf.groups.ingroup
        data['inconf'] = bs.traf.cd.inconf
        data['tcpamax'] = bs.traf.cd.tcpamax
        data['rpz'] = bs.traf.cd.rpz
        data['nconf_cur'] = len(bs.traf.cd.confpairs_unique)
        data['nconf_tot'] = len(bs.traf.cd.confpairs_all)
        data['nlos_cur'] = len(bs.traf.cd.lospairs_unique)
        data['nlos_tot'] = len(bs.traf.cd.lospairs_all)
        data['trk']        = bs.traf.trk
        data['vs']         = bs.traf.vs
        data['vmin']       = bs.traf.perf.vmin
        data['vmax']       = bs.traf.perf.vmax

        # Transition level as defined in traf
        data['translvl']   = bs.traf.translvl

        # Send casmachthr for route visualization
        data['casmachthr']    = aero.casmach_thr

        # ASAS resolutions for visualization. Only send when evaluated
        data['asastas']  = bs.traf.cr.tas
        data['asastrk']  = bs.traf.cr.trk

        # Aircraft (group) color
        if self.custacclr:
            data['custacclr'] = self.custacclr
        if self.custgrclr:
            data['custgrclr'] = self.custgrclr

        return data

    def send_route_data(self):
        ''' Send route data to client(s) '''
        # Case 1: A route is selected by one or more specific clients
        if self.client_route:
            for sender, acid in self.client_route.items():
                self._sendrte(sender, acid)
            # Check if there are other senders and also a scenario-selected route
            if self.route_all:
                remclients = bs.sim.clients.difference(self.client_route.keys())
                for sender in remclients:
                    self._sendrte(sender, self.route_all)
        # Case 2: only a route selected from scenario file:
        # Broadcast the same route to everyone
        elif self.route_all:
            self._sendrte(b'*', self.route_all)
        
    def _sendrte(self, sender, acid):
        ''' Local shorthand function to send route. '''
        data               = dict()
        data['acid']       = acid
        idx   = bs.traf.id2idx(acid)
        if idx >= 0:
            route          = bs.traf.ap.route[idx]
            data['iactwp'] = route.iactwp

            # We also need the corresponding aircraft position
            data['aclat']  = bs.traf.lat[idx]
            data['aclon']  = bs.traf.lon[idx]

            data['wplat']  = route.wplat
            data['wplon']  = route.wplon

            data['wpalt']  = route.wpalt
            data['wpspd']  = route.wpspd

            data['wpname'] = route.wpname

        self.pub_route.send_replace((sender or b'C'), **data)


#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from distutils.core import setup, Extension
import numpy as np

ext_modules = [Extension('_cgeo', sources=['_cgeo.cpp'])]

setup(name='_cgeo', version='1.0', include_dirs=[np.get_include()],
      ext_modules=ext_modules)



""" BlueSky implementation of a timer that can periodically trigger functions."""
import time
from math import remainder
from bluesky.core.signal import Signal


class Timer:
    """ A timer can be used to periodically (wall-time) trigger functions."""

    # Data that the wall clock needs to keep
    __timers__ = []

    def __init__(self, interval: int):
        super().__init__()
        self.timeout  = Signal()
        self.interval = interval * 1000000
        self.t_next   = (time.time_ns() // self.interval + 1) * self.interval

        # Append self to known timers
        Timer.__timers__.append(self)

    @classmethod
    def update_timers(cls):
        ''' Update all timers. '''
        tcur = time.time_ns()
        for timer in cls.__timers__:
            if tcur >= timer.t_next:
                timer.timeout.emit()
                timer.t_next = (tcur // timer.interval + 1) * timer.interval



from time import time, gmtime, strftime
import numpy as np
import matplotlib.pyplot as plt
from math import degrees
import collections
from collections import defaultdict
import itertools as IT

import bluesky as bs
from bluesky.tools import geo
from bluesky.tools.misc import tim2txt
from bluesky.tools.aero import *
from bluesky import settings

# Register settings defaults
settings.set_variable_defaults(log_path='output')

"""
    This module seems to work as follows:
    It looks like this used to be a submodule of the traffic module
    Now it is a full module under the sim package

    Apparently the creation of this module also called into life the concept of a research area.
    A research area is something which is not specific to the metrics module and could be used by different modules.
    However, the area command in the stack module saves data in the metric instance.
    If the area function and metric module are to be seperatly used, then they should be untangled.

    Classes:
    Metric, metric_Area, metric_CoCa, metric_HB
    Each has a constructor
    Requirements for instance creation:
        Metric: -
        metric_Area: -
        metric_CoCa: regions
        metric_HB: area

    Called from:
        metric_Area: Metric
        metric_CoCa: Metric
        metric_HB: Metric

    Passed as argument to:
        metric_Area: metric_CoCa, metric_HB (only the .cellArea instance)
        metric_CoCa: -
        metric_HB: -

    Structure:
        metric object created with Metric class constructor
        metric.metric object is of the form (metric_CoCa, metric_HB)
            So the metric instance has two sub instances,
            one of type metric_CoCa and one of type metric_HB
"""
class metric_Area():
    def __init__(self):
        self.lat = 55.5
        self.lon = 1.7
        self.fll = 8500
        self.flu = 41500
        self.bearingS = 180
        self.bearingE = 90
        self.deltaFL = 3000
        self.distance = 20
        self.ncells = 18
        self.nlevels = 12

        self.regions = np.array([0,0,0])

    def addbox(self,lat,lon):

        lat_0 = lat
        lat_00 = lat
        lon_0 = lon
        londiviser = 1
        for i in range(1,self.ncells+1):
            for j in range(1,self.ncells+1):
                for k in range(self.fll,self.flu+self.deltaFL,self.deltaFL):
                    box = np.array([lat,lon,k])
                    self.regions = np.vstack([self.regions,box])

                if i == 1:

                    lat,lon = geo.qdrpos(lat,lon,self.bearingE,self.distance)
                    lat = degrees(lat)
                    lon = degrees(lon)
                    londiviser = (lon - lon_0) / self.ncells
                else:
                    lat,lon = geo.qdrpos(lat,lon,self.bearingE,self.distance)
                    lat = degrees(lat)
                    lon = lon_0 + londiviser * j

            lat_0 = lat_00
            lat,lon = geo.qdrpos(lat_0,lon_0,self.bearingS,self.distance*i)
            lat = degrees(lat)
            lon = degrees(lon)
            lat_0 = lat

        return

    def cellArea(self):
        point1 = [self.regions[0,0],self.regions[0,1]]
        point2 = [self.regions[self.ncells*self.nlevels-1,0],self.regions[self.ncells*self.nlevels-1,1]]
        point3 = [self.regions[(self.ncells-1)*self.ncells*self.nlevels,0],self.regions[(self.ncells-1)*self.ncells*self.nlevels,1]]
        point4 = [self.regions[self.ncells*self.ncells*self.nlevels-1,0],self.regions[self.ncells*self.ncells*self.nlevels-1,1]]
        self.cellarea = np.array([point4,point2,point1,point3])
        #print self.cellarea
        return self.cellarea

    def makeRegions(self):
        lat = self.lat
        lon = self.lon

        self.addbox(lat,lon)
        self.regions = np.delete(self.regions, (0), axis=0)

        return self.regions


    def area_of_polygon(self,x, y):
        area = 0.0
        for i in range(-1, len(x) - 1):
            area += x[i] * (y[i + 1] - y[i - 1])
        return area / 2.0

    def centroid_of_polygon(self,points):
        area = self.area_of_polygon(*list(zip(*points)))

        result_x = 0
        result_y = 0
        N = len(points)

        points = IT.cycle(points)

        x1, y1 = next(points)
        for i in range(N):
            x0, y0 = x1, y1
            x1, y1 = next(points)
            cross = (x0 * y1) - (x1 * y0)
            result_x += (x0 + x1) * cross
            result_y += (y0 + y1) * cross
        result_x /= (area * 6.0)
        result_y /= (area * 6.0)
        return (result_x, result_y)


    def FIR_circle(self, fir_number):
        fir_lat = []
        fir_lon = []
        fir = []

        fir_lat.append(bs.navdb.fir[fir_number][1])
        fir_lon.append(bs.navdb.fir[fir_number][2])
        fir.append((fir_lat[-1],fir_lon[-1]))

        fir = fir[0]
        fir = list(zip(fir[0],fir[1]))
        fir_centroid = self.centroid_of_polygon(fir)

        return fir_centroid

class metric_CoCa():

    def __init__(self,regions):


        self.region = regions
        self.oldaircraft = np.zeros((1000,1), dtype = [('callsign','|S10'),('cellnumber',int), ('time',int),('totaltime',int)])
        self.newaircraft = np.zeros((1000,1), dtype = [('callsign','|S10'),('cellnumber',int), ('time',int),('totaltime',int)])
        # self.cells = np.zeros((self.region.nlevels*self.region.ncells*self.region.ncells,1), dtype = [('cellnumber',int),('interactions',int),('ntraf',int)])
        # for i in range(0,len(self.cells)):
        #    self.cells['cellnumber'][i] = i + 1
        #        plt.close()

        self.numberofcells = self.region.ncells*self.region.ncells*self.region.nlevels

        names = []
        for i in range(0,self.numberofcells):
            names.append("cell"+str(i))

        formats = []
        for i in range(0,self.numberofcells):
            formats.append("|S10")

        ndtype = {'names':names, 'formats':formats}
        self.cells = np.zeros((500,6), dtype = ndtype)
        self.resettime = 5 #seconds
        self.deltaresettime = self.resettime
        self.iteration = 0

        formats = []
        for i in range(0,self.numberofcells):
            formats.append(float)

        ndtype = {'names':names, 'formats':formats}
        oneday = 86400 # second in one day
        numberofrows = oneday / self.resettime
        numberofrows = 3
        self.precocametric = np.zeros((numberofrows,5), dtype = ndtype)
        self.cocametric = np.zeros((numberofrows,6), dtype = ndtype)
        plt.ion()
        self.ntraf = 0

        # plt.colorbar()
        # self.plotntraf,= plt.plot([], [])
        # self.plotbar, = plt.bar([],[])
        return

    def findCell(self,cells,lat,lon,fl):
        i = 0
        j = 0
        k = 0

        for i in range(0,len(cells),self.region.ncells*self.region.nlevels):
            if (cells[0,0]) <= lat < (cells[0,0]+0.6):
                break
            if (cells[i,0] < lat < cells[0,0] and cells[i,0] > cells[-1,0]) :
                break
            else:
                i = -10000

        if i > -1 :

            for j in range(0,self.region.ncells*self.region.nlevels,self.region.nlevels):

                if cells[i+j,1] > lon and lon < (cells[-1,1]+0.6) and lon > cells[0,1]:
                    j = j - self.region.nlevels
                    break
                if cells[i+j,1]+0.6 > lon and lon < (cells[-1,1]+0.6) and lon > cells[0,1]:
                    break
                else:
                    j = -10000

            if j > - 1:
                for k in range(0,self.region.nlevels,1):
                    if cells[i+j+k,2] > fl and fl < (cells[-1,2]+self.region.deltaFL) and fl > cells[0,2]:
                        k = k -1
                        break
                    else:
                        k = -10000

        if (i+j+k) < 0:
            i=-1
            j=0
            k=0

        return i+j+k


    # def update_line(self,ntraf,t):
    #     if t < 0.1:

    #        self.__init__()

    #     t = int(t)
    #     self.plotntraf.set_xdata(np.append(self.plotntraf.get_xdata(), t))
    #     self.plotntraf.set_ydata(np.append(self.plotntraf.get_ydata(), ntraf))
    #     plt.plot(t,ntraf,'b--o')
    #     ax = plt.gca()
    #     ax.relim()
    #     ax.autoscale_view()
    #     return

    # def update_bar(self,trafcell,t):
    #    if t < 0.1:
    #        self.__init__
    #    t = int(t)
    #    self.plotbar.set_xdata(np.append(self.plotbar.get_xdata(), t))
    #    self.plotbar.set_ydata(np.append(self.plotbar.get_ydata(), trafcell))
    #    plt.plot(t,ntraf,'b--o')
    #    ax = plt.gca()
    #    ax.relim()
    #    ax.autoscale_view()
    #    return

    # def plot_interactions(self):
    #      plotcells = np.sort(self.cells, axis = 0, order='interactions')[-3:]
    #      label = np.vstack(plotcells['cellnumber'])
    #      x = np.arange(len(label))
    #      y1 = np.vstack(plotcells['interactions'])
    #      colLabels=("Interactions","")
    #      nrows, ncols = len(x)+1, len(colLabels)
    #      hcell, wcell = 0.3, 0.5
    #      hpad, wpad = 0, 0.5
    #      fig1=plt.figure(num = 1, figsize=(ncols*wcell+wpad, nrows*hcell+hpad))
    #      ax = fig1.add_subplot(111)
    #      ax.axis('off')
    #      ax.table(cellText=y1,colLabels=colLabels, rowLabels = label ,loc='center')
    #      plt.bar(x,y)
    #      plt.xticks(x,str(label))
    #      plt.show()
    #     return

    # def cell_interactions(self,cellN):
    #    # Interactions
    #    itemscount = np.array(collections.Counter(cellN).items())
    #    if len(itemscount) > 0:
    #        for number in range(0,self.region.ncells*self.region.ncells*self.region.nlevels):
    #            j = np.where(itemscount[:,0] == number)
    #            if np.size(j) == 0:
    #                self.cells['interactions'][number] = 0
    #            else:
    #                self.cells['interactions'][number] = itemscount[j,1]*(itemscount[j,1]-1)
    #        self.plot_interactions()
    #    return


    # def celltime(self,time):
    #     for i in range(0,len(self.newaircraft)):
    #         j = np.where(self.oldaircraft['callsign'] == self.newaircraft['callsign'][i])[0]
    #         if np.size(j) == 1:
    #             if self.oldaircraft['cellnumber'][j] == self.newaircraft['cellnumber'][i]:
    #                 self.newaircraft['time'][i] = time - self.oldaircraft['totaltime'][i]
    #             else:
    #                 self.newaircraft['totaltime'][i] = time - self.oldaircraft['totaltime'][i]
    #                 self.cells['ntraf'][i] = self.cells['ntraf'][i] + 1
    #     self.oldaircraft = self.newaircraft
    #     return self.newaircraft


    def cellPlot(self):
        cell = [floor(x/12) for x in bs.traf.cell]
        count = collections.Counter(cell)
        count = np.array(list(count.items()))

        flcells = count
        if np.size(count)>0:
            flcells = count[:,0]

        z = np.array([0])
        for number in range(0,self.region.ncells*self.region.ncells):
            i = np.where(flcells == (number))

            if np.size(i) == 0:
                z = np.append(z,0)
            else:
                i = i[0]
                z = np.append(z,count[i,1])

        z = np.delete(z, (0), axis=0)
        zdata = np.reshape(z,(-1,self.region.ncells))
        fig = plt.figure(1)
        ax = fig.add_subplot(1, 1, 1)
        ax.imshow(zdata, interpolation='nearest')
        plt.show()
        return

    def applyMetric(self):
        for i in range(0,self.numberofcells):
            name = 'cell'+str(i)

            l = self.iteration

            times = []
            headings = []
            speeds = []
            vspeeds = []
            actimes = []

            for j in range(0,len(self.cells[name])):
                if self.cells[name][j][1] != "":
                    times.append(float(self.cells[name][j][1]))
                    headings.append(float(self.cells[name][j][2]))
                    speeds.append(float(self.cells[name][j][3]))
                    vspeeds.append(float(self.cells[name][j][4]))
                    actimes.append(float(self.cells[name][j][1]))

            indices = np.argsort(times)
            times.sort()

            headings = [headings[z] for z in indices]
            speeds = [speeds[y] for y in indices]
            vspeeds = [vspeeds[x] for x in indices]
            actimes = [actimes[w] for w in indices]


            for w in range(0,len(vspeeds)):
                if vspeeds[w] <= 500 and vspeeds[w] >= (-500):
                    vspeeds[w] = 0
                elif vspeeds[w] > 500:
                    vspeeds[w] = 1
                elif vspeeds[w] < (-500):
                    vspeeds[w] = -1



            self.precocametric[name][l][0] = (sum(times)/self.deltaresettime)

            acinteractions = []
            spdinteractions = []
            hdginteractions = []
            vspdinteractions = []

            if len(times) > 1:
                for k in range(0,len(times)):
                    aircraft = len(times)

                    time = times[0]/self.deltaresettime
                    actime = actimes[0]/self.deltaresettime
                    acinteractions.append(aircraft*(aircraft-1)*(actime**aircraft))


                    counter = 0
                    for t in range(0,1):
                        for u in range(t+1,len(speeds)):
                            if abs(speeds[t]-speeds[u]) > 35:
                                counter = counter + 1
                    spdinteractions.append(2*counter*(time**(counter+1)))


                    counter = 0
                    for t in range(0,1):
                        for u in range(t+1,len(headings)):
                            if abs(headings[t]-headings[u]) > 20:
                                counter = counter + 1
                    hdginteractions.append(2*counter*(time**(counter+1)))


                    counter = 0
                    for t in range(0,1):
                        for u in range(t+1,len(vspeeds)):
                            if vspeeds[t] != vspeeds[u]:
                                counter = counter + 1
                    vspdinteractions.append(2*counter*(time**(counter+1)))

                    for x in range(1,len(actimes)):
                        actimes[x] = actimes[x] - actimes[0]

                    del actimes[0]
                    del times[0]
                    del vspeeds[0]
                    del speeds[0]
                    del headings[0]

                self.precocametric[name][l][1] = sum(acinteractions)
                self.precocametric[name][l][2] = sum(spdinteractions)
                self.precocametric[name][l][3] = sum(hdginteractions)
                self.precocametric[name][l][4] = sum(vspdinteractions)

                self.cocametric[name][l][1] = self.precocametric[name][l][1] / self.precocametric[name][l][0]
                self.cocametric[name][l][2] = self.precocametric[name][l][2] / self.precocametric[name][l][0]
                self.cocametric[name][l][3] = self.precocametric[name][l][3] / self.precocametric[name][l][0]
                self.cocametric[name][l][4] = self.precocametric[name][l][4] / self.precocametric[name][l][0]
                self.cocametric[name][l][0] = self.cocametric[name][l][1] * (self.cocametric[name][l][2] + self.cocametric[name][l][3] + self.cocametric[name][l][4])

        print("Iteration number: "+str(self.iteration+1))
        print("Reset time = "+str(self.resettime))
        return


    def reset(self):
        names = []
        for i in range(0,self.numberofcells):
            names.append("cell"+str(i))

        formats = []
        for i in range(0,self.numberofcells):
            formats.append("|S10")

        ndtype = {'names':names, 'formats':formats}
        self.cells = np.zeros((500,6), dtype = ndtype)

        return

    def AircraftCell(self,cells,time):
        if floor(time) >= self.resettime:
            bs.sim.hold()
            self.reset()
            self.resettime = self.resettime + self.deltaresettime
            self.iteration = self.iteration + 1
            # filedata = bs.resource(settings.log_path) / "coca_20120727-78am-1hour.npy"
            # self.cellPlot(traf)
            # np.save(filedata,self.cocametric)
            bs.sim.op()

        bs.traf.cell = []

        for i in range(bs.traf.ntraf):
            lat = bs.traf.lat[i]
            lon = bs.traf.lon[i]
            fl = bs.traf.alt[i]/ft
            cellN = self.findCell(cells,lat,lon,fl)

            if cellN > 0:
                bs.traf.cell = np.append(bs.traf.cell, cellN)
                name = 'cell'+str(cellN)

                index = np.where(bs.traf.id[i] == self.cells[name][:,[0]])[0]
                if len(index) != 1:

                    j = 0
                    for j in range(0,len(self.cells[name])):
                        if self.cells[name][j][0] == "":
                            break
                    self.cells[name][j][0] = bs.traf.id[i]
                    self.cells[name][j][1] = time
                    self.cells[name][j][2] = bs.traf.ahdg[i]
                    self.cells[name][j][3] = eas2tas(bs.traf.selspd[i],bs.traf.selalt[i])/kts
                    self.cells[name][j][4] = bs.traf.selvs[i]/fpm
                    self.cells[name][j][5] = time
                if len(index) == 1:
                    createtime = float(self.cells[name][index[0]][5])
                    self.cells[name][index[0]][1] = str(time - createtime)

                # self.newaircraft['callsign'][i] = bs.traf.id[i]
                # self.newaircraft['cellnumber'][i] =
        return


class metric_HB():

    def __init__(self,area):
        self.initiallat = area[3][0]
        self.initiallon = area[3][1]
        self.dist_range = 5.0 #nm

        self.alt_range = 1000.0 #ft
        self.t_cpa = 0
        self.dist_cpa = 0
        self.spd = np.array([])
        self.lat = np.array([])
        self.lon = np.array([])
        self.pos = np.array([])
        self.trk = 0

        self.alt_dif = 0
        self.alt = 0
        self.id = []

        self.complexity = defaultdict(lambda:defaultdict(int))
        self.rel_trk = np.array([])
        self.step = -1
        self.id_previous = []
        self.headings = []
        self.headings_previous = np.array([])
        self.doubleconflict = 0
        self.ntraf = 0
        self.compl_ac = 0
        self.time_lookahead = 1800 #seconds

        self.selected_area = ([area[0][0],area[0][1]],[area[1][0],area[1][1]],[area[2][0],area[2][1]],[area[3][0],area[3][1]])

        return

    def selectTraffic(self):

        traf_selected_lat = np.array([])
        traf_selected_lon = np.array([])
        traf_selected_alt = np.array([])
        traf_selected_tas = np.array([])
        traf_selected_trk = np.array([])
        traf_selected_ntraf = 0
        # RECTANGLE AREA
        # for i in range(0,bs.traf.ntraf):
        #     if nx.pnpoly(bs.traf.lat[i],bs.traf.lon[i],self.selected_area) == 1:
        #         traf_selected_lat = np.append(traf_selected_lat,bs.traf.lat[i])
        #         traf_selected_lon = np.append(traf_selected_lon,bs.traf.lon[i])
        #         traf_selected_alt = np.append(traf_selected_alt,bs.traf.alt[i])
        #         traf_selected_tas = np.append(traf_selected_tas,bs.traf.tas[i])
        #         traf_selected_trk = np.append( traf_selected_trk,bs.traf.trk[i])
        #         traf_selected_ntraf = traf_selected_ntraf + 1

        # CIRCLE AREA (FIR Circle)
        for i in range(0,bs.traf.ntraf):

            dist = latlondist(bs.sim.metric.fir_circle_point[0],\
                              bs.sim.metric.fir_circle_point[1],\
                              bs.traf.lat[i],bs.traf.lon[i])

            if  dist/nm < bs.sim.metric.fir_circle_radius:
                traf_selected_lat = np.append(traf_selected_lat,bs.traf.lat[i])
                traf_selected_lon = np.append(traf_selected_lon,bs.traf.lon[i])
                traf_selected_alt = np.append(traf_selected_alt,bs.traf.alt[i])
                traf_selected_tas = np.append(traf_selected_tas,bs.traf.tas[i])
                traf_selected_trk = np.append( traf_selected_trk,bs.traf.trk[i])
                traf_selected_ntraf = traf_selected_ntraf + 1


        return traf_selected_lat,traf_selected_lon,traf_selected_alt,traf_selected_tas,traf_selected_trk,traf_selected_ntraf


    def applymetric(self):
        time1 = time()
        bs.sim.hold()
        self.doubleconflict = 0
        # relative pos x and pos y
        self.step = self.step + 1
        self.pos = np.array([])
        self.lat = np.array([])
        self.lon = np.array([])

        self.id = []
        self.alt_dif = 0

        traf_selected_lat,traf_selected_lon,traf_selected_alt,traf_selected_tas,traf_selected_trk,traf_selected_ntraf = self.selectTraffic()


        [self.rel_trk, self.pos] = geo.qdrdist_matrix(self.initiallat,self.initiallon,np.mat(traf_selected_lat),np.mat(traf_selected_lon))
        # self.lat = np.append(self.lat,traf.lat)
        # self.lon = np.append(self.lon,traf.lon)
        self.id = bs.traf.id

        # Position x and y wrt to initial position
        self.pos = np.mat(self.pos)
        anglex = np.cos(np.radians(90-self.rel_trk))
        angley = np.sin(np.radians(90-self.rel_trk))

        self.posx = np.mat(np.array(self.pos) * np.array(anglex)) #nm
        self.posy = np.mat(np.array(self.pos) * np.array(angley)) #nm

        self.lat = traf_selected_lat
        self.lon = traf_selected_lon

        self.alt = np.mat(traf_selected_alt/ft)
        self.spd = traf_selected_tas/nm #nm/s
        self.trk = traf_selected_trk
        self.ntraf = traf_selected_ntraf

        self.alt_dif = self.alt-self.alt.T
        # Vectors CPA_dist and CPA_time
        self.apply_twoCircleMethod()
        time2 = time()
        print("Time to Complete Calculation: " + str(time2-time1))

        bs.sim.op()
        return


    def rel_matrixs(self):
        self.alt_dif = self.alt-self.alt.T
        # speeds
        hdgx = np.cos(np.radians(90-self.trk))
        hdgy = np.sin(np.radians(90-self.trk))

        spdu = np.mat(self.spd * hdgx.T).T #nm/s
        spdv = np.mat(self.spd * hdgy.T).T #nm/s

        # distances pos and spd
        distx = np.array(self.posx.T - self.posx) #nm
        disty = np.array(self.posy.T - self.posy) #nm
        distu = (np.array(spdu.T - spdu)) #nm/s
        distv = (np.array(spdv.T - spdv)) #nm/s

        # predicted time to CPA
        self.t_cpa = -(distu*distx+distv*disty)/      \
         (distu*distu+distv*distv+np.array(np.eye(distu[:,0].size)))

        # predicted distance to CPA
        relcpax = self.t_cpa*np.array(spdu.T)
        relcpay = self.t_cpa*np.array(spdv.T)
        cpax = self.posx.T + relcpax
        cpay = self.posy.T + relcpay
        distcpax = np.array(cpax-cpax.T)
        distcpay = np.array(cpay-cpay.T)
        self.dist_cpa = (distcpax**2+distcpay**2)**0.5

        return

    def apply_altfilter(self,S0):
        condition = abs(self.alt_dif)<self.alt_range
        # self.t_cpa = np.where(condition,self.t_cpa, np.nan)
        # self.dist_cpa = np.where(condition,self.dist_cpa, np.nan)
        S0 = np.where(condition,S0,np.nan)
        return S0

    def apply_distfilter(self,H0):
        condition = self.dist_cpa<self.dist_range*3
        self.dist_cpa = np.where(condition,self.dist_cpa,np.nan)
        self.t_cpa = np.where(condition,self.t_cpa,np.nan)
        H0 = np.where(condition,H0,np.nan)
        return H0

    def apply_timefilter(self):
        condition = self.t_cpa>0#(self.t_cpa<(self.time_range+20) * (self.t_cpa>0))
        self.t_cpa = np.where(condition,self.t_cpa,np.nan)
        self.dist_cpa = np.where(condition,self.dist_cpa,np.nan)
        return

    def apply_before_filter(self,S0,Va):
        Vb = Va.T
        Va_Vb = np.add(np.abs(Va),np.abs(Vb))
        condition1 = S0>0
        condition2 = np.divide(S0,Va_Vb)>self.time_lookahead #seconds
        condition = np.multiply(condition1,condition2)
        condition = np.invert(condition)
        S0 = np.where(condition,S0,np.nan)
        return S0

    def merge(self,times):
            if len(times) > 0:
                saved = list(times[0])

                for st, en in sorted([(t) for t in times]):
                    if st <= saved[1]:
                        saved[1] = max(saved[1], en)
                    else:
                        yield list(saved)
                        saved[0] = st
                        saved[1] = en
                yield list(saved)
            else:
                yield list(times)

    def apply_twoCircleMethod(self):
        Va = np.mat(self.spd)
        Ha = np.radians(self.trk)

        Vb = np.add(Va,0.0000001)
        VaVa = np.multiply(Va,Va)

        Hb = Ha

        [H0,S0] = geo.qdrdist_matrix(np.mat(self.lat),np.mat(self.lon),np.mat(self.lat),np.mat(self.lon))
        S0 = np.where(S0 > 0, S0, np.nan)

        S0 = self.apply_before_filter(S0,Va)
        S0 = self.apply_altfilter(S0)

        H0 = np.radians(H0.T)

        R_S0 = np.divide(self.dist_range,S0)
        arcsin = np.arcsin(R_S0)

        ha_new11,ha_new21,ha_new12,ha_new22,t1d1,t1d2,t2d1,t2d2 = self.calc_angles(Vb,Hb,VaVa,H0,arcsin,S0)

        R_S0 = None
        arcsin = None

        ha_1 = np.degrees(ha_new11)
        ha_3 = np.degrees(ha_new21)

        ha_2 = np.degrees(ha_new12)
        ha_4 = np.degrees(ha_new22)

        t1 = t1d1
        t2 = t1d2
        t3 = t2d1
        t4 = t2d2

        ha_new11 = None
        ha_new21 = None
        ha_new12 = None
        ha_new22 = None
        t1d1 = None
        t1d2 = None
        t2d1 = None
        t2d2 = None

        ha_1,ha_2,ha_3,ha_4,t1,t2,t3,t4 = self.conditions(ha_1,ha_2,ha_3,ha_4,t1,t2,t3,t4,Va,Vb,Ha,Hb)

        ## Condition where S0 < self.dist_range
        condition = np.multiply(S0<self.dist_range,S0>0)

        ac_angles = {}
        ac_score = {}
        for k in range(0,self.ntraf):
            ac_angles[str(k)] = []
            ac_score[str(k)] = 0
            for l in range(0,self.ntraf):
                if not np.isnan(ha_1[l,k]) and not np.isnan(ha_2[l,k]):
                    ac_angles[str(k)].append((ha_1[l,k],ha_2[l,k]))
                if not np.isnan(ha_3[l,k]) and not np.isnan(ha_4[l,k]):
                    ac_angles[str(k)].append((ha_3[l,k],ha_4[l,k]))

            ac_angles[str(k)] = sorted(ac_angles[str(k)])
            ac_angles[str(k)] = list(self.merge(ac_angles[str(k)]))
            if len(ac_angles[str(k)][0]) > 0:
                for z in range(0,len(ac_angles[str(k)])):
                    ac_angle180min = ((self.trk[k]+180-90)%360-180)
                    ac_angle180max = ((self.trk[k]+180+90)%360-180)
                    ac_angles_st180 = ((ac_angles[str(k)][z][0]+180)%360-180)
                    ac_angles_en180 = ((ac_angles[str(k)][z][-1]+180)%360-180)

                    ac_angle360min = (self.trk[k]+360-90)%360
                    ac_angle360max = (self.trk[k]+360+90)%360
                    ac_angles_st360 = (ac_angles_st180+360)%360
                    ac_angles_en360 = (ac_angles_en180+360)%360

                    if ac_angle180min<90 and ac_angle180min>-90:
                        if ac_angles_st180 < ac_angle180min:
                            ac_angles[str(k)][z][0] = ac_angle180min
                    else:
                        if ac_angles_st360 < ac_angle360min:
                            ac_angles[str(k)][z][0] = ac_angle180min

                    if ac_angle180max < 90 and ac_angle180max > -90:
                        if ac_angles_en180 > ac_angle180max:
                            ac_angles[str(k)][z][-1] = ac_angle180max
                    else:
                        if ac_angles_en360 > ac_angle360max:
                            ac_angles[str(k)][z][-1] = ac_angle180max

                    if (ac_angles_st180 < ac_angle180min and ac_angles_en180 < ac_angle180min) or (ac_angles_st360 < ac_angle360min and ac_angles_en360 < ac_angle360min):
                        ac_angles[str(k)][z] = [np.nan,np.nan]

                    # Complexity Score
                    if ac_angles[str(k)][z][-1]<90 and ac_angles[str(k)][z][-1]>-90:
                        ac_score[str(k)] = ac_score[str(k)] + (ac_angles[str(k)][z][-1]-ac_angles[str(k)][z][0])/180
                    else:
                        ac_score[str(k)] = ac_score[str(k)] + (ac_angles[str(k)][z][-1]+360-ac_angles[str(k)][z][0])/180

            if True in condition[k]:
                ac_score[str(k)] = 1

            if np.isnan(ac_score[str(k)]):
                ac_score[str(k)] = 0

        ac_totalscore =  sum(ac_score.values())

        self.complexity[self.step][0] = ac_totalscore #/ self.ntraf
        self.complexity[self.step][1] = ac_totalscore / max(1,self.ntraf)

        print("Complexity per Aircraft: " + str(self.complexity[self.step][1]))
        return


    def calc_angles(self,Vb,Hb,VaVa,H0,arcsin,S0):
        wx = np.multiply(Vb,np.sin(Hb))
        wy = np.multiply(Vb,np.cos(Hb))

        wxwx = np.multiply(wx,wx)
        wywy = np.multiply(wy,wy)
        wxwx_wywy = np.add(wxwx,wywy)

        a = np.subtract(wxwx_wywy.T,VaVa)

        H0_arcsin = np.subtract(H0,arcsin)
        xc1 = np.multiply(S0,np.sin(H0_arcsin))
        yc1 = np.multiply(S0,np.cos(H0_arcsin))

        xc1xc1 = np.multiply(xc1,xc1)
        yc1yc1 = np.multiply(yc1,yc1)

        a = np.where(a!=0,a,np.nan)
        a = np.add(a,0.00000000001)

        b1 = np.add(2*np.multiply(xc1,wx.T),2*np.multiply(yc1,wy.T))
        c1 = np.add(xc1xc1,yc1yc1)
        b1b1 = np.multiply(b1,b1)
        d1 = np.subtract(b1b1,4*np.multiply(a,c1))

        c1 = None
        b1b1 = None
        xc1xc1 = None
        yc1yc1 = None
        H0_arcsin = None

        a_2 = np.multiply(a,2)
        conditiond1 = d1<0
        conditiond1 = np.invert(conditiond1)
        d1 = np.where(conditiond1,d1,np.nan)
        t01d1 = np.divide(np.subtract(-b1,np.sqrt(d1)),a_2)
        t02d1 = np.divide(np.add(-b1,np.sqrt(d1)),a_2)

        t1d1 = np.minimum(t01d1,t02d1)
        t2d1 = np.maximum(t01d1,t02d1)

        xpt1d1 = np.add(xc1,np.multiply(wx.T,t1d1))
        xpt2d1 = np.add(xc1,np.multiply(wx.T,t2d1))

        ypt1d1 = np.add(yc1,np.multiply(wy.T,t1d1))
        ypt2d1 = np.add(yc1,np.multiply(wy.T,t2d1))

        xc1 = None
        yc1 = None

        H0_arcsin = np.add(H0,arcsin)
        xc2 = np.multiply(S0,np.sin(H0_arcsin))
        yc2 = np.multiply(S0,np.cos(H0_arcsin))

        xc2xc2 = np.multiply(xc2,xc2)
        yc2yc2 = np.multiply(yc2,yc2)

        wxT = wx.T
        wyT = wy.T
        xc2_wx = np.multiply(xc2,wxT)
        yc2_wy = np.multiply(yc2,wyT)
        xc2_wx2 = np.multiply(xc2_wx,2)
        yc2_wy2 = np.multiply(yc2_wy,2)
        b2 = np.add(xc2_wx2,yc2_wy2)
        c2 = np.add(xc2xc2,yc2yc2)
        b2b2 = np.multiply(b2,b2)
        d2 = np.subtract(b2b2,4*np.multiply(a,c2))

        c2 = None
        b2b2 = None
        xc2xc2 = None
        yc2yc2 = None
        H0_arcsin = None

        wxwx = None
        wywy = None
        wxwx_wywy = None

        conditiond2 = d2<0
        conditiond2 = np.invert(conditiond2)
        d2 = np.where(conditiond2,d2,np.nan)

        t01d2 = np.divide(np.subtract(-b2,np.sqrt(d2)),a_2)
        t02d2 = np.divide(np.add(-b2,np.sqrt(d2)),a_2)
        t1d2 = np.minimum(t01d2,t02d2)
        t2d2 = np.maximum(t01d2,t02d2)

        t01d1 = None
        t02d1 = None
        t01d2 = None
        t02d2 = None

        d1 = None
        d2 = None

        xpt1d2 = np.add(xc2,np.multiply(wx.T,t1d2))
        xpt2d2 = np.add(xc2,np.multiply(wx.T,t2d2))

        ypt1d2 = np.add(yc2,np.multiply(wy.T,t1d2))
        ypt2d2 = np.add(yc2,np.multiply(wy.T,t2d2))

        xc2 = None
        yc2 = None
        wx = None
        wy = None

        ha_new11 = np.arctan2(xpt1d1,ypt1d1)
        ha_new21 = np.arctan2(xpt2d1,ypt2d1)

        ha_new12 = np.arctan2(xpt1d2,ypt1d2)
        ha_new22 = np.arctan2(xpt2d2,ypt2d2)

        return ha_new11,ha_new21,ha_new12,ha_new22,t1d1,t1d2,t2d1,t2d2

    def conditions(self,ha_1,ha_2,ha_3,ha_4,t1,t2,t3,t4,Va,Vb,Ha,Hb):
        t1_nan = t1>0
        t2_nan = t2>0
        t3_nan = t3>0
        t4_nan = t4>0

        ha_1 = np.where(t1_nan,ha_1,np.nan)
        ha_2 = np.where(t2_nan,ha_2,np.nan)
        ha_3 = np.where(t3_nan,ha_3,np.nan)
        ha_4 = np.where(t4_nan,ha_4,np.nan)

        t1 = np.where(t1_nan,t1,np.nan)
        t2 = np.where(t2_nan,t2,np.nan)
        t3 = np.where(t3_nan,t3,np.nan)
        t4 = np.where(t4_nan,t4,np.nan)

        # condition: Va < Vb and all + t's
        Va_Vb = Va < Vb
        t1_t2 = np.multiply(t1_nan,t2_nan)
        t3_t4 = np.multiply(t3_nan,t4_nan)
        t_allplus = np.multiply(t1_t2,t3_t4)
        condition = np.multiply(Va_Vb,t_allplus)
        condition = np.invert(condition)
        ha_3new = np.where(condition,ha_3,ha_4)
        t3new = np.where(condition,t3,t4)
        ha_4new = np.where(condition,ha_4,ha_3)
        t4new = np.where(condition,t4,t3)

        # condition Va < Vb and t1,t3 negatif
        t1_neg = np.invert(t1_nan)
        t3_neg = np.invert(t3_nan)
        t1_t3_neg = np.multiply(t1_neg,t3_neg)
        condition = np.multiply(Va_Vb,t1_t3_neg)
        condition = np.invert(condition)
        ha_3 = np.where(condition,ha_3new,ha_4new)
        t3 = np.where(condition,t3new,t4new)
        ha_4 = np.where(condition,ha_4new,ha_2)
        t4 = np.where(condition,t4new,t2)

        # condition Va < Vb and t2,t4 negatif
        t2_neg = np.invert(t2_nan)
        t4_neg = np.invert(t4_nan)
        t2_t4_neg = np.multiply(t2_neg,t4_neg)
        condition = np.multiply(Va_Vb,t2_t4_neg)
        condition = np.invert(condition)
        ha_2 = np.where(condition,ha_2,ha_3)
        t2 = np.where(condition,t2,t3)

        # TBD More than 90-degree turns!
        Ha = np.degrees(Ha)
        Hb = np.degrees(Hb)

        # Lookahead time
        t1_lht = t1 > self.time_lookahead
        t2_lht = t2 > self.time_lookahead
        t1_t2_lht = np.multiply(t1_lht,t2_lht)
        t1_t2_lht = np.invert(t1_t2_lht)

        t3_lht = t3 > self.time_lookahead
        t4_lht = t4 > self.time_lookahead
        t3_t4_lht = np.multiply(t3_lht,t4_lht)
        t3_t4_lht = np.invert(t3_t4_lht)

        ha_1 = np.where(t1_t2_lht,ha_1,np.nan)
        ha_2 = np.where(t1_t2_lht,ha_2,np.nan)
        ha_3 = np.where(t3_t4_lht,ha_3,np.nan)
        ha_4 = np.where(t3_t4_lht,ha_4,np.nan)

        t1 = np.where(t1_t2_lht,t1,np.nan)
        t2 = np.where(t1_t2_lht,t2,np.nan)
        t3 = np.where(t3_t4_lht,t3,np.nan)
        t4 = np.where(t3_t4_lht,t4,np.nan)

        return ha_1,ha_2,ha_3,ha_4,t1,t2,t3,t4

    def saveData(self):
        acid = np.array(self.id).reshape(-1,).tolist()
        lat = np.array(self.lat).reshape(-1,).tolist()
        lon = np.array(self.lon).reshape(-1,).tolist()
        compl = np.array(self.compl_ac).reshape(-1,).tolist()
        alt =  np.array(self.alt).reshape(-1,).tolist()
        spd = np.array(self.spd).reshape(-1,).tolist()
        trk = np.array(self.trk).reshape(-1,).tolist()
        ntraf = np.repeat(np.array(self.ntraf),len(trk))
        ntraf = np.array(ntraf).reshape(-1,).tolist()

        data = zip(acid,lat,lon,alt,spd,trk,ntraf,compl)

        step = str(self.step).zfill(3)
        fname = bs.resource(settings.log_path) / "Metric-HB" / f"{step}-BlueSky.csv"
        f = csv.writer(open(fname, "wb"))
        for row in data:
            f.writerow(row)
        return

    def complexity_plot(self):
        if self.step == 0:
            self.plot_complexity,= plt.plot([], [])
            self.plot_complexity2, = plt.plot([],[])
            plt.ion()

        t = int(self.step)
        self.plot_complexity.set_xdata(np.append(self.plot_complexity.get_xdata(), t))
        self.plot_complexity.set_ydata(np.append(self.plot_complexity.get_ydata(), self.complexity[self.step][0]))
        self.plot_complexity2.set_xdata(np.append(self.plot_complexity2.get_xdata(), t))
        self.plot_complexity2.set_ydata(np.append(self.plot_complexity2.get_ydata(), self.complexity[self.step][1]))

        plt.subplot(2,1,1)
        plt.plot(t,self.complexity[self.step][0],'b--o--')
        plt.subplot(2,1,2)
        plt.plot(t,self.complexity[self.step][1],'g--o--')
        plt.draw()
        return


        #    def heading_range(self,i,ac1_conflict,ac2_conflict,lat0_ac1,lon0_ac1,lat0_ac2,lon0_ac2):
        #
        #        h_range = np.arange(int(self.trk[ac1_conflict[i]])-90,int(self.trk[ac1_conflict[i]])+90,1)
        #        h_range = np.append(h_range,self.trk[ac2_conflict[i]])
        #
        #
        #        hdgx = np.cos(np.radians(90-h_range))
        #        hdgy = np.sin(np.radians(90-h_range))
        #
        #        spd = np.array([])
        #        posy = np.array([])
        #        posx = np.array([])
        #        for l in range(0,len(h_range)-1):
        #            posy = np.append(posy,self.posy[ac1_conflict[i]])
        #            posx = np.append(posx,self.posx[ac1_conflict[i]])
        #            spd = np.append(spd,self.spd[ac1_conflict[i]])
        #
        #        spd = np.append(spd,self.spd[ac2_conflict[i]])
        #
        #        spdu = np.mat(spd * hdgx).T #nm/s
        #        spdv = np.mat(spd * hdgy).T #nm/s
        #        posy = np.append(posy,self.posy[ac2_conflict[i]])
        #        posx = np.append(posx,self.posx[ac2_conflict[i]])
        #        posx = np.mat(posx).T
        #        posy = np.mat(posy).T
        #
        #        ## distances pos and spd
        #        distx = np.array(posx - posx.T) #nm
        #        disty = np.array(posy - posy.T) #nm
        #        distu = (np.array(spdu - spdu.T)) #nm/s
        #        distv = (np.array(spdv - spdv.T)) #nm/s
        #
        #        ## predicted time to CPA
        #        t_heading = -(distu*distx+distv*disty)/      \
        #         (distu*distu+distv*distv+np.array(np.eye(distu[:,0].size)))
        #
        #        t_heading = self.apply_timefilter(t_heading)
        #
        #        relcpax = t_heading*np.array(spdu.T)
        #        relcpay = t_heading*np.array(spdv.T)
        #        cpax = posx.T + relcpax
        #        cpay = posy.T + relcpay
        #        distcpax = np.array(cpax-cpax.T)
        #        distcpay = np.array(cpay-cpay.T)
        #        distcpa = (distcpax**2+distcpay**2)**0.5
        #
        #        n_headings = np.where(distcpa>self.dist_range)
        #        headings = np.array([])
        #
        #        for z in range(0,len(n_headings[0])/2):
        #            headings = np.append(headings,h_range[n_headings[0][z]])
        #
        #
        #
        #
        #        #check if flightid is same, so multiple conflicts per id
        #        if self.id[ac1_conflict[i]] == self.id_previous:
        #            headings = np.intersect1d(self.headings_previous,headings)
        #            n_fault_headings = len(h_range)-len(headings)
        #            ratio = float(float(n_fault_headings)/float(len(h_range)))
        #            add_complexity = [self.id[ac1_conflict[i]],ratio,headings]
        #            self.complexity[self.step][i-self.doubleconflict-1] = add_complexity
        #            self.doubleconflict = self.doubleconflict + 1
        #        else:
        #            n_fault_headings = len(h_range)-len(headings)
        #            ratio = float(float(n_fault_headings)/float(len(h_range)))
        #            add_complexity = [self.id[ac1_conflict[i]],ratio,headings]
        #            self.complexity[self.step][i-self.doubleconflict] = add_complexity
        #
        #
        #        self.id_previous = self.id[ac1_conflict[i]]
        #        self.headings_previous = headings
        #
        #        return
        #
        #
        #
        #    def apply_precisetime(self,range_t,delta_t,heading_range):
        #
        #        ac_conflict = np.where(self.t_cpa>0)
        #        conflicts = np.size(ac_conflict)/2
        #
        #        if conflicts > 0:
        #            ac1_conflict = ac_conflict[0]
        #            ac2_conflict = ac_conflict[1]
        #
        #
        #            t_cpa_precise = np.array([])
        #            plat_ac1 = np.array([])
        #            plat_ac2 = np.array([])
        #            plon_ac1 = np.array([])
        #            plon_ac2 = np.array([])
        #
        #            for i in range(0,conflicts):
        #                lat0_ac1 = self.lat[ac1_conflict[i]]
        #                lat0_ac2 = self.lat[ac2_conflict[i]]
        #                lon0_ac1 = self.lon[ac1_conflict[i]]
        #                lon0_ac2 = self.lon[ac2_conflict[i]]
        #
        #                # DETAILED CPA
        ##                t_cpa_ac1ac2 = self.t_cpa[ac1_conflict[i]][ac2_conflict[i]]
        ##                t_cpa_range = np.arange(t_cpa_ac1ac2-range_t,t_cpa_ac1ac2+range_t,delta_t)
        ##                dist_ac1 = t_cpa_precise*self.spd[ac1_conflict[i]]*nm
        ##                dist_ac2 = t_cpa_precise*self.spd[ac2_conflict[i]]*nm
        ##
        ##
        ##                plat_ac1 = np.append(plat_ac1,lat0_ac1 + \
        ##                             np.degrees(dist_ac1*np.cos(np.radians(self.trk[ac1_conflict[i]]))/Rearth))
        ##                plat_ac2 = np.append(plat_ac2,lat0_ac2 + \
        ##                             np.degrees(dist_ac2*np.cos(np.radians(self.trk[ac2_conflict[i]]))/Rearth))
        ##                plon_ac1 = np.append(plon_ac1,lon0_ac1 + \
        ##                                   np.degrees(dist_ac1*np.sin(np.radians(self.trk[ac1_conflict[i]])) \
        ##                                     /np.cos(np.radians(lat0_ac1))/Rearth))
        ##                plon_ac2 = np.append(plon_ac2,lon0_ac2 + \
        ##                                   np.degrees(dist_ac2*np.sin(np.radians(self.trk[ac2_conflict[i]])) \
        ##                                       /np.cos(np.radians(lat0_ac2))/Rearth))
        ##
        ##                dist_ac1ac2 = np.array([])
        ##                for j in range(0,np.size(t_cpa_range)):
        ##                    dist_ac1ac2 = np.append(dist_ac1ac2,(latlondist(plat_ac1[j],plon_ac1[j],plat_ac2[j],plon_ac2[j]))/nm)
        ##
        ##                index_new_t_cpa = np.argmin(dist_ac1ac2)
        ##                self.dist_cpa[ac1_conflict[i]][ac2_conflict[i]] = dist_ac1ac2[index_new_t_cpa]
        ##
        ##                self.t_cpa[ac1_conflict[i]][ac2_conflict[i]] = t_cpa_precise[index_new_t_cpa]
        #
        #                if heading_range == 1:
        #                    self.heading_range(i,ac1_conflict,ac2_conflict,lat0_ac1,lon0_ac1,lat0_ac2,lon0_ac2)
        #
        #        return

    def apply_heading_range(self):
        Va = np.mat(self.spd)
        Ha = np.radians(self.trk)
        Ha_len = len(Ha)
        Ha = np.repeat(Ha,Ha_len)

        Ha = Ha.reshape(Ha_len,Ha_len)

        Vb = Va
        Hb = Ha

        R = self.dist_range

        [H0,S0] = geo.qdrdist_matrix(np.mat(self.lat),np.mat(self.lon),np.mat(self.lat),np.mat(self.lon))

        S0 = self.apply_before_filter(S0,Va)

        H0 = np.radians(H0.T)
        R_S0 = np.divide(R,S0)
        arcsin = np.arcsin(R_S0)
        Hr_new1 = H0 - arcsin
        Hr_new2 = H0 + arcsin
        Vb_Va = np.divide(Vb.T,Va)
        Hb_Hr_sin1 = np.sin(Hr_new1 - Hb)
        mp1 = np.multiply(Vb_Va,Hb_Hr_sin1)
        mp1 = (mp1 + 1)%2 - 1
        Hb_Hr_sin2 = np.sin(Hr_new2 - Hb)
        mp2 = np.multiply(Vb_Va,Hb_Hr_sin2)
        mp2 = (mp2 + 1)%2 - 1

        Ha_new11 = Hr_new1 - np.arcsin(mp1)
        Ha_new12 = Hr_new1 - np.arcsin(np.pi-mp1)
        Ha_new21 = Hr_new2 - np.arcsin(mp2)
        Ha_new22 = Hr_new2 - np.arcsin(np.pi-mp2)

        Ha_new11 = np.degrees(Ha_new11)
        Ha_new12 = np.degrees(Ha_new12)
        Ha_new21 = np.degrees(Ha_new21)
        Ha_new22 = np.degrees(Ha_new22)

        Ha_new11 = (Ha_new11+180)%360-180
        Ha_new12 = (Ha_new12+180)%360-180
        Ha_new21 = (Ha_new21+180)%360-180
        Ha_new22 = (Ha_new22+180)%360-180

        # Check for more than 90-degree headings, TBD
        Hdeg = np.degrees(Ha.T)
        Heading_1 = np.subtract(Hdeg,90)
        Heading_2 = np.add(Hdeg,90)

        # Heading_new11
        condition11a = Ha_new11 < Heading_1
        condition11b = S0>0
        condition11 = np.multiply(condition11a,condition11b)
        condition11 = np.invert(condition11)
        Ha_new11 = np.where(condition11,Ha_new11,Heading_1)

        # Heading_new21
        condition21a = Ha_new21 > Heading_2
        condition21 = np.multiply(condition21a,condition11b)
        condition21 = np.invert(condition21)
        Ha_new21 = np.where(condition21,Ha_new21,Heading_2)

        # Condition where Va>Vb and angle_rel does not include in H0_limits
        va_sxa = np.multiply(Va,np.sin(Ha))
        vb_sxb = va_sxa
        va_cxa = np.multiply(Va,np.cos(Ha))
        vb_cxb = va_cxa
        angle_rel = np.arctan2(np.subtract(va_sxa.T,vb_sxb),np.subtract(va_cxa.T,vb_cxb))
        angle_rel = np.degrees(angle_rel)
        angle_rel = (angle_rel+180)%360-180

        condition1 = Va.T > Vb
        H0_lim1 = np.degrees(H0) - 90
        H0_lim2 = np.degrees(H0) + 90
        H0_lim1 = (H0_lim1+180)%360-180
        H0_lim2 = (H0_lim2+180)%360-180
        condition2a = angle_rel < H0_lim1
        condition2b = angle_rel > H0_lim2

        condition2 = np.multiply(condition2a,condition2b)
        condition = np.multiply(condition1,condition2)
        condition = np.invert(condition)

        Ha_new11 = np.where(condition,Ha_new11,np.nan)
        Ha_new12 = np.where(condition,Ha_new12,np.nan)
        Ha_new21 = np.where(condition,Ha_new21,np.nan)
        Ha_new22 = np.where(condition,Ha_new22,np.nan)

        # S0_S0 = np.multiply(S0,S0)
        # d_sep = self.dist_range**2
        # aaa = np.multiply(2,np.multiply(S0,self.dist_range))
        # d_min = H0
        # condition = np.abs(np.subtract(H0,x_rel)) < np.radians(90)
        # d_min = np.where(condition,d_min,np.multiply(H0,np.sin(np.subtract(H0,x_rel))))
        # cos_b = np.arcsin(np.divide(d_min,self.dist_range)) - np.abs(np.subtract(H0,x_rel))
        # mp = np.multiply(aaa,cos_b)
        # upper = np.sqrt(S0_S0 + d_sep - mp)
        # Va_Va = np.multiply(Va,Va)
        # Vb_Vb = np.multiply(Vb,Vb)
        # va_vb = np.multiply(Va,Vb)
        # a_va_vb = np.multiply(2,va_vb)
        # mult = np.multiply(a_va_vb,np.cos(np.subtract(Ha.T,Hb)))
        # Va_Va_Vb_Vb = np.add(Va_Va,Vb_Vb)

        # v_rel = np.sqrt(np.subtract(Va_Va_Vb_Vb,mult))
        # print v_rel
        # t_fls = np.divide(upper,v_rel)
        # print t_fls

        diff_11_21 = np.subtract(Ha_new11,Ha_new21)
        diff_11_21 = np.abs((diff_11_21 + 180) % 360 - 180)

        diff_12_22 = np.subtract(Ha_new12,Ha_new22)
        diff_12_22 = np.abs((diff_12_22 + 180)  % 360 - 180)

        # now combine into 1 complexity score
        Compl_1 = np.divide(diff_11_21,360)
        Compl_2 = np.divide(diff_12_22,360)
        Compl_12ac = np.vstack((Compl_1,Compl_2))
        Compl_ac = np.nansum(Compl_12ac, axis = 0)

        # Condition where S0 < self.dist_range
        # condition = np.multiply(S0<self.dist_range,S0>0)
        # condition = np.invert(condition)
        # Compl_ac = np.where(condition,Compl_ac,0.5)

        Compl_ac = np.nansum(Compl_ac)

        self.complexity[self.step][0] = Compl_ac #/ self.ntraf
        self.complexity[self.step][1] = Compl_ac / self.ntraf

        self.complexity_plot()
        return


class Metric():
    """
    Metric class definition : traffic metrics

    Methods:
        Metric()                :  constructor

        update()                : add a command to the command stack
        close()                 : close file

    Created by  : Jacco M. Hoekstra (TU Delft)

    GitHub Metric, full collaborator MICHON-BS!!
    """

    def __init__(self):
        # Create metrics file
        # fname = settings.log_path + \
        #     strftime("%Y-%m-%d-%H-%M-%S-BlueSky-Metrics.txt", gmtime())
        # self.file = open(fname,"w")

        # Write header
#        self.write(0.0,"Header info tbd")

        # Last time for which Metrics.update was called
        self.t0 = -9999   # force first time call

        # Set time interval in seconds
        self.dt = 1  # [seconds]

        self.name = ("CoCa-Metric","HB-Metric","Delete AC")
        self.metric_number = -1
        self.fir_circle_point = [0.,0.]
        self.fir_circle_radius = 0.
        self.fir_number = 0
        self.metricstime = 0
        self.tbegin = 0

        self.metric_Area = metric_Area()

        self.cells = self.metric_Area.makeRegions()

        self.cellarea = self.metric_Area.cellArea()
        self.metric = (metric_CoCa(self.metric_Area),metric_HB(self.cellarea))

        return

    def toggle(self, flag, dt=None):
        """ Toggle metrics module from stack """
        if type(flag) == bool and not flag:
            # Can toggle metrics off with 'OFF'
            self.metric_number = -1
        else:
            # Otherwise select a metric to run
            if flag <= 0:
                return True, "Metrics OFF"

            if flag > len(self.name):
                return False, "No such metric"

            if not bs.traf.area == "Circle":
                return False, "First define AREA for metric"

            if dt:
                self.dt = dt

            self.metric_number = flag - 1

            return True, "Activated %s (%d), dt=%.2f" % (self.name[self.metric_number], flag, self.dt)

    def write(self,t,line):
        """
        Write: can be called from anywhere traf.metric.write( txt )
        Adds time stamp and ';'
        """
        self.file.write(tim2txt(t)+";"+line+chr(13)+chr(10))
        return

    def update(self):
        #check if configured and there is actual traffic
        if self.metric_number == -1 or bs.traf.ntraf < 1:
            return

        """Update: to call for regular logging & runtime analysis"""
        # Only do something when time is there
        if abs(bs.sim.simt-self.t0)<self.dt:
            return
        self.t0 = bs.sim.simt  # Update time for scheduler
        if self.metricstime == 0:
            self.tbegin = bs.sim.simt
            self.metricstime = 1
            print("METRICS STARTED")
            # FIR_circle(bs.navdb,self.fir_number)
            # cmd.stack("AREA "+str(self.cellarea[2][0])+","+str(self.cellarea[2][1])+ \
            #   ","+str(self.cellarea[0][0])+","+str(self.cellarea[0][1]))

        # A lot of smart Michon-code here, probably using numpy arrays etc.
        if bs.sim.simt >= 0:
            if self.metric_number == 0:
                self.metric[self.metric_number].AircraftCell(self.cells,bs.sim.t-self.tbegin)
            elif self.metric_number == 1:
                self.metric[self.metric_number].applymetric()

        print("Number of Aircraft in Research Area (FIR):" + str(self.metric[self.metric_number].ntraf))

        deleteAC = []
        for i in range(0,bs.traf.ntraf):
            if bs.traf.selvs[i] <= 0 and (bs.traf.selalt[i]/ft) < 750 and bs.traf.selspd[i] < 300:
                deleteAC.append(bs.traf.id[i])

            elif bs.traf.selvs[i] <=0 and (bs.traf.selalt[i]/ft) < 10.:
                deleteAC.append(bs.traf.id[i])

            if bs.traf.selvs[i] <=0 and bs.traf.selspd[i] < 10.:
                deleteAC.append(bs.traf.id[i])

        for i in range(0,len(deleteAC)):
            bs.traf.delete(deleteAC[i])

        # Heartbeat for test
        self.write(bs.sim.simt,"NTRAF;"+str(bs.traf.ntraf))
        return

    def plot(self):
        # Pause simulation
        bs.sim.hold()

        # Open a plot window attached to a command?
        #    plot, showplot and other matplotlib commands

        # Continue simulation
        bs.sim.op()
        return



""" This module provides PerfBase, the base class for aircraft
    performance implementations.
"""
import numpy as np
from bluesky import settings
from bluesky.core import Entity, timed_function
from bluesky.stack import command


settings.set_variable_defaults(performance_dt=1.0)


class PerfBase(Entity, replaceable=True):
    """Base class for BlueSky aircraft performance implementations."""

    def __init__(self):
        super().__init__()
        with self.settrafarrays():
            # --- fixed parameters ---
            self.actype = np.array([], dtype=str)  # aircraft type
            self.Sref = np.array([])  # wing reference surface area [m^2]
            self.engtype = np.array([])  # integer, aircraft.ENG_TF...

            # --- dynamic parameters ---
            self.mass = np.array([])  # effective mass [kg]
            self.phase = np.array([])
            self.cd0 = np.array([])
            self.k = np.array([])
            self.bank = np.array([])
            self.thrust = np.array([])  # thrust
            self.drag = np.array([])  # drag
            self.fuelflow = np.array([])  # fuel flow

            # Envelope limits per aircraft
            self.hmax = np.array([])  # Flight ceiling [m]
            self.vmin = np.array([])  # Minimum operating speed [m/s]
            self.vmax = np.array([])  # Maximum operating speed [m/s]
            self.vsmin = np.array([])  # Maximum descent speed [m/s]
            self.vsmax = np.array([])  # Maximum climb speed [m/s]
            self.axmax = np.array([])  # Max/min acceleration [m/s2]

    def create(self, n):
        super().create(n=n)
        # Set wide default limits, so that no envelope limiting occurs
        # when the actual performance model used doesn't support it
        self.axmax[-n:] = 2.0  # Default acceleration limit is 2 m/s2
        self.hmax[-n:] = 1e6
        self.vmin[-n:] = -1e6
        self.vmax[-n:] = 1e6
        self.vsmin[-n:] = -1e6
        self.vsmax[-n:] = 1e6


    @timed_function(name="performance", dt=settings.performance_dt, hook="preupdate")
    def update(self, dt=settings.performance_dt):
        """implement this method"""
        pass

    def limits(self, intent_v, intent_vs, intent_h, ax):
        """implement this method"""
        return intent_v, intent_vs, intent_h

    def currentlimits(self):
        """implement this method"""
        # Get current kinematic performance envelop of all aircraft
        pass

    @command(name="ENG")
    def engchange(self, acid: "acid", engine_id: "txt" = ""):
        """Specify a different engine type for aircraft 'acid'"""
        return (
            False,
            "The currently selected performance model doesn't support engine changes.",
        )

    @command(name="PERFSTATS", aliases=("PERFINFO", "PERFDATA"))
    def show_performance(self, acid: "acid"):
        """Show aircraft perfromance parameters for aircraft 'acid'"""
        return (
            False,
            "The currently selected performance model doesn't provide this function.",
        )

    @staticmethod
    @command(name="PERF")
    def setmethod(name: "txt" = ""):
        """Select a Performance implementation."""
        # Get a dict of all registered Performance models
        methods = PerfBase.derived()
        names = ["OFF" if n == "PERFBASE" else n for n in methods]

        if not name:
            curname = (
                "OFF"
                if PerfBase.selected() is PerfBase
                else PerfBase.selected().__name__
            )
            return (
                True,
                f"Current Performance model: {curname}"
                + f'\nAvailable performance models: {", ".join(names)}',
            )
        # Check if the requested method exists
        if name == "OFF":
            PerfBase.select()
            return True, "Performance model turned off."
        method = methods.get(name, None)
        if method is None:
            return (
                False,
                f"{name} doesn't exist.\n"
                + f'Available performance models: {", ".join(names)}',
            )

        # Select the requested method
        method.select()
        return True, f"Selected {method.__name__} as performance model."
